try:
    import unittest2 as unittest
except ImportError:
    import unittest
import nose2
from unittest.mock import MagicMock, patch, mock_open, call
import os
from lxml import etree
import sys
import json

from ncclient.manager import Manager, make_device_handler
from ncclient.transport import SSHSession
import ncclient.transport.errors as NcErrors
from ncclient.operations import RPCError, TimeoutExpiredError

from jnpr.junos.facts.swver import version_info
from jnpr.junos import Device
from jnpr.junos.exception import RpcError
from jnpr.junos import exception as EzErrors
from jnpr.junos.console import Console

__author__ = "Rick Sherman, Nitin Kumar, Stacy Smith"
__credits__ = "Jeremy Schulman"

if sys.version < "3":
    builtin_string = "__builtin__"
else:
    builtin_string = "builtins"

facts = {
    "domain": None,
    "hostname": "firefly",
    "ifd_style": "CLASSIC",
    "version_info": version_info("15.1X46-D15.3"),
    "2RE": False,
    "serialnumber": "aaf5fe5f9b88",
    "fqdn": "firefly",
    "virtual": True,
    "switch_style": "NONE",
    "version": "12.1X46-D15.3",
    "HOME": "/cf/var/home/rick",
    "srx_cluster": False,
    "model": "FIREFLY-PERIMETER",
    "RE0": {
        "status": "Testing",
        "last_reboot_reason": "Router rebooted after a " "normal shutdown.",
        "model": "FIREFLY-PERIMETER RE",
        "up_time": "6 hours, 29 minutes, 30 seconds",
    },
    "vc_capable": False,
    "personality": "SRX_BRANCH",
}


class Test_MyTemplateLoader(unittest.TestCase):
    def setUp(self):
        from jnpr.junos.device import _MyTemplateLoader

        self.template_loader = _MyTemplateLoader()

    @patch(builtin_string + ".filter")
    def test_temp_load_get_source_filter_false(self, filter_mock):
        filter_mock.return_value = []
        try:
            self.template_loader.get_source(None, None)
        except Exception as ex:
            import jinja2

            self.assertEqual(type(ex), jinja2.exceptions.TemplateNotFound)

    @patch("jnpr.junos.device.os.path")
    def test_temp_load_get_source_filter_true(self, os_path_mock):
        # cant use @patch here as with statement will have exit
        m = mock_open()
        with patch(builtin_string + ".open", m, create=True):
            self.template_loader.get_source(None, None)


class TestDevice(unittest.TestCase):
    @patch("ncclient.manager.connect")
    def setUp(self, mock_connect):
        mock_connect.side_effect = self._mock_manager

        self.dev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        self.dev.open()

    @patch("ncclient.operations.session.CloseSession.request")
    def tearDown(self, mock_session):
        self.dev.close()

    def test_new_console_return(self):
        dev = Device(
            host="1.1.1.1",
            user="test",
            password="password123",
            port=23,
            gather_facts=False,
        )
        self.assertTrue(isinstance(dev, Console))

    @patch("jnpr.junos.device.netconf_ssh")
    def test_device_ConnectAuthError(self, mock_manager):
        mock_manager.connect.side_effect = NcErrors.AuthenticationError
        self.assertRaises(EzErrors.ConnectAuthError, self.dev.open)

    @patch("jnpr.junos.device.netconf_ssh")
    def test_device_ConnectRefusedError(self, mock_manager):
        mock_manager.connect.side_effect = NcErrors.SSHError
        self.assertRaises(EzErrors.ConnectRefusedError, self.dev.open)

    @patch("jnpr.junos.device.netconf_ssh")
    @patch("jnpr.junos.device.datetime")
    def test_device_ConnectTimeoutError(self, mock_datetime, mock_manager):
        mock_manager.connect.side_effect = NcErrors.SSHError(
            "Could not open socket to 1.1.1.1:830"
        )
        from datetime import timedelta, datetime

        currenttime = datetime.now()
        mock_datetime.datetime.now.side_effect = [
            currenttime,
            currenttime + timedelta(minutes=4),
        ]
        self.assertRaises(EzErrors.ConnectTimeoutError, self.dev.open)

    @patch("jnpr.junos.device.netconf_ssh")
    @patch("jnpr.junos.device.datetime")
    def test_device_diff_err_message(self, mock_datetime, mock_manager):
        NcErrors.SSHError.message = "why are you trying :)"
        mock_manager.connect.side_effect = NcErrors.SSHError
        from datetime import timedelta, datetime

        currenttime = datetime.now()
        mock_datetime.datetime.now.side_effect = [
            currenttime,
            currenttime + timedelta(minutes=4),
        ]
        self.assertRaises(EzErrors.ConnectError, self.dev.open)

    @patch("jnpr.junos.device.netconf_ssh")
    def test_device_ConnectUnknownHostError(self, mock_manager):
        import socket

        mock_manager.connect.side_effect = socket.gaierror
        self.assertRaises(EzErrors.ConnectUnknownHostError, self.dev.open)

    @patch("jnpr.junos.device.netconf_ssh")
    def test_device_other_error(self, mock_manager):
        mock_manager.connect.side_effect = TypeError
        self.assertRaises(EzErrors.ConnectError, self.dev.open)

    def test_device_probe_error(self):
        mock_probe = MagicMock()
        mock_probe.return_value = None
        self.dev.probe = mock_probe

        def fn():
            self.dev.open(auto_probe=1)

        self.assertRaises(EzErrors.ProbeError, fn)

    def test_device_property_logfile_isinstance(self):
        mock = MagicMock()
        with patch(builtin_string + ".open", mock):
            if sys.version > "3":
                builtin_file = "io.TextIOWrapper"
            else:
                builtin_file = builtin_string + ".file"
            with patch(builtin_file, MagicMock):
                handle = open("filename", "r")
                self.dev.logfile = handle
                self.assertEqual(self.dev.logfile, handle)

    def test_device_host_mand_param(self):
        self.assertRaises(
            ValueError, Device, user="test", password="password123", gather_facts=False
        )

    def test_device_property_logfile_close(self):
        self.dev._logfile = MagicMock()
        self.dev._logfile.close.return_value = 0
        self.dev.logfile = None
        self.assertFalse(self.dev._logfile)

    def test_device_property_logfile_exception(self):
        try:
            self.dev.logfile = True
        except Exception as ex:
            self.assertEqual(type(ex), ValueError)

    @patch("jnpr.junos.Device.execute")
    def test_device_uptime(self, mock_execute):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        mock_execute.side_effect = self._mock_manager
        self.assertEqual(localdev.uptime, 14234)

    def test_device_master_is_master(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        localdev.facts._cache["current_re"] = [
            "re1",
            "master",
            "node",
            "fwdd",
            "member",
            "pfem",
        ]
        self.assertEqual(localdev.master, True)

    def test_device_master_gnf_is_master(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        localdev.facts._cache["current_re"] = ["gnf1-re0", "gnf1-master"]
        localdev.facts._cache["hostname_info"] = {
            "bsys-re0": "foo",
            "bsys-re1": "foo1",
            "gnf1-re0": "bar",
            "gnf1-re1": "bar1",
        }
        self.assertEqual(localdev.master, True)

    def test_device_master_is_backup(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        localdev.facts._cache["current_re"] = ["re0", "backup"]
        self.assertEqual(localdev.master, False)

    def test_device_master_gnf_is_backup(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        localdev.facts._cache["current_re"] = ["gnf1-re1", "gnf1-backup"]
        localdev.facts._cache["hostname_info"] = {
            "bsys-re0": "foo",
            "bsys-re1": "foo1",
            "gnf1-re0": "bar",
            "gnf1-re1": "bar1",
        }
        self.assertEqual(localdev.master, False)

    def test_device_master_is_re0_only(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        localdev.facts._cache["2RE"] = False
        localdev.facts._cache["RE_hw_mi"] = False
        localdev.facts._cache["current_re"] = ["re0"]
        self.assertEqual(localdev.master, True)

    def test_device_master_is_multi_chassis_non_master1(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        localdev.facts._cache["2RE"] = True
        localdev.facts._cache["current_re"] = [
            "lcc1-re1",
            "member1-re1",
            "lcc1-backup",
            "member1-backup",
        ]
        self.assertEqual(localdev.master, False)

    def test_device_master_is_multi_chassis_non_master2(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        localdev.facts._cache["2RE"] = True
        localdev.facts._cache["current_re"] = [
            "lcc1-re0",
            "member1-re0",
            "lcc1-master",
            "member1-master",
            "member1",
        ]
        self.assertEqual(localdev.master, False)

    def test_device_master_is_none1(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        localdev.facts._cache["current_re"] = None
        self.assertEqual(localdev.master, None)

    def test_device_master_is_none2(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        localdev.facts._cache["2RE"] = True
        localdev.facts._cache["current_re"] = ["foo", "bar"]
        self.assertEqual(localdev.master, None)

    @patch("jnpr.junos.device.warnings")
    def test_device_master_is_old_facts(self, mock_warn):
        localdev = Device(
            host="1.1.1.1",
            user="test",
            password="password123",
            fact_style="old",
            gather_facts=False,
        )
        mock_warn.assert_has_calls(
            [
                call.warn(
                    "fact-style old will be removed " "in a future release.",
                    RuntimeWarning,
                )
            ]
        )
        self.assertEqual(localdev.master, None)

    def test_device_master_setter(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        with self.assertRaises(RuntimeError):
            localdev.master = "foo"

    def test_device_re_name_is_re0(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        localdev.facts._cache["current_re"] = ["re0", "backup"]
        localdev.facts._cache["hostname_info"] = {"re0": "tapir", "re1": "tapir1"}
        self.assertEqual(localdev.re_name, "re0")

    def test_device_re_name_is_lcc_re1(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        localdev.facts._cache["current_re"] = [
            "lcc1-re1",
            "member1-re1",
            "lcc1-backup",
            "member1-backup",
        ]
        localdev.facts._cache["hostname_info"] = {"re0": "mj1"}
        self.assertEqual(localdev.re_name, "lcc1-re1")

    def test_device_re_name_is_re0_only(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        localdev.facts._cache["current_re"] = ["foo"]
        localdev.facts._cache["hostname_info"] = {"re0": "mj1"}
        self.assertEqual(localdev.re_name, "re0")

    def test_device_re_name_is_bsys_re0(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        localdev.facts._cache["current_re"] = ["re0"]
        localdev.facts._cache["hostname_info"] = {"bsys-re0": "foo"}
        self.assertEqual(localdev.re_name, "bsys-re0")

    def test_device_re_name_is_none1(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        localdev.facts._cache["current_re"] = None
        self.assertEqual(localdev.re_name, None)

    def test_device_re_name_is_none2(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        localdev.facts._cache["current_re"] = [
            "re1",
            "master",
            "node",
            "fwdd",
            "member",
            "pfem",
        ]
        localdev.facts._cache["hostname_info"] = None
        self.assertEqual(localdev.re_name, None)

    @patch("jnpr.junos.device.warnings")
    def test_device_re_name_is_old_facts(self, mock_warn):
        localdev = Device(
            host="1.1.1.1",
            user="test",
            password="password123",
            fact_style="old",
            gather_facts=False,
        )
        mock_warn.assert_has_calls(
            [
                call.warn(
                    "fact-style old will be removed " "in a future release.",
                    RuntimeWarning,
                )
            ]
        )
        self.assertEqual(localdev.re_name, None)

    def test_device_re_name_setter(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        with self.assertRaises(RuntimeError):
            localdev.re_name = "foo"

    def test_device_repr(self):
        localdev = Device(
            host="1.1.1.1", user="test", password="password123", gather_facts=False
        )
        self.assertEqual(repr(localdev), "Device(1.1.1.1)")

    def test_device_local(self):
        Device.ON_JUNOS = True
        localdev = Device()
        self.assertEqual(localdev._hostname, "localhost")
        Device.ON_JUNOS = False

    @patch("jnpr.junos.device.os")
    @patch(builtin_string + ".open")
    @patch("paramiko.config.SSHConfig.lookup")
    def test_device__sshconf_lkup(self, mock_paramiko, open_mock, os_mock):
        os_mock.path.exists.return_value = True
        self.dev._sshconf_lkup()
        mock_paramiko.assert_called_once_with("1.1.1.1")

    @patch("jnpr.junos.device.os")
    @patch(builtin_string + ".open")
    @patch("paramiko.config.SSHConfig.lookup")
    def test_device__sshconf_lkup_def(self, mock_paramiko, open_mock, os_mock):
        os_mock.path.exists.return_value = True
        self.dev._ssh_config = "/home/rsherman/.ssh/config"
        self.dev._sshconf_lkup()
        mock_paramiko.assert_called_once_with("1.1.1.1")

    @patch("paramiko.config.SSHConfig.lookup")
    def test_device__sshconf_lkup_sock_fd(self, mock_paramiko):
        self.dev2 = Device(sock_fd=6)
        self.dev2._sshconf_lkup()
        self.assertEqual(self.dev2._sshconf_lkup(), None)

    @patch("os.path.expanduser")
    def test_device__sshconf_lkup_path_not_exists(self, mock_path):
        mock_path.return_value = "/home/test"
        self.assertEqual(self.dev._sshconf_lkup(), None)

    @patch("ncclient.manager.connect")
    @patch("jnpr.junos.Device.execute")
    def test_device_open(self, mock_connect, mock_execute):
        with patch("jnpr.junos.utils.fs.FS.cat") as mock_cat:
            mock_cat.return_value = """

    domain jls.net

            """
            mock_connect.side_effect = self._mock_manager
            mock_execute.side_effect = self._mock_manager
            self.dev2 = Device(host="2.2.2.2", user="test", password="password123")
            self.dev2.open()
            self.assertEqual(self.dev2.connected, True)

    @patch("ncclient.manager.connect")
    @patch("jnpr.junos.Device.execute")
    def test_device_open_with_bind_addr(self, mock_connect, mock_execute):
        with patch("jnpr.junos.utils.fs.FS.cat") as mock_cat:
            mock_cat.return_value = """

    domain jls.net

            """
            mock_connect.side_effect = self._mock_manager
            mock_execute.side_effect = self._mock_manager
            self.dev2 = Device(
                host="2.2.2.2", user="test", password="password123", bind_addr="1.1.1.1"
            )
            self.dev2.open()
            self.assertEqual(self.dev2.connected, True)

    @patch("ncclient.manager.connect")
    @patch("jnpr.junos.Device.execute")
    def test_device_open_with_look_for_keys_False(self, mock_connect, mock_execute):
        with patch("jnpr.junos.utils.fs.FS.cat") as mock_cat:
            mock_cat.return_value = """

    domain jls.net

            """
            mock_connect.side_effect = self._mock_manager
            mock_execute.side_effect = self._mock_manager
            self.dev2 = Device(
                host="2.2.2.2", user="test", password="password123", look_for_keys=False
            )
            self.dev2.open()
            self.assertEqual(self.dev2.connected, True)

    @patch("ncclient.manager.connect")
    @patch("jnpr.junos.Device.execute")
    def test_device_open_with_look_for_keys_True(self, mock_connect, mock_execute):
        with patch("jnpr.junos.utils.fs.FS.cat") as mock_cat:
            mock_cat.return_value = """

    domain jls.net

            """
            mock_connect.side_effect = self._mock_manager
            mock_execute.side_effect = self._mock_manager
            self.dev2 = Device(
                host="2.2.2.2", user="test", password="password123", look_for_keys=True
            )
            self.dev2.open()
            self.assertEqual(self.dev2.connected, True)

    @patch("ncclient.manager.connect")
    @patch("jnpr.junos.Device.execute")
    def test_device_open_with_hostkey_verify_True(self, mock_connect, mock_execute):
        with patch("jnpr.junos.utils.fs.FS.cat") as mock_cat:
            mock_cat.return_value = """

    domain jls.net

            """
            mock_connect.side_effect = self._mock_manager
            mock_execute.side_effect = self._mock_manager
            self.dev2 = Device(
                host="2.2.2.2", user="test", password="password123", hostkey_verify=True
            )
            self.dev2.open()
            self.assertEqual(self.dev2.connected, True)

    @patch("ncclient.manager.connect")
    @patch("jnpr.junos.Device.execute")
    def test_device_open_with_hostkey_verify_False(self, mock_connect, mock_execute):
        with patch("jnpr.junos.utils.fs.FS.cat") as mock_cat:
            mock_cat.return_value = """

    domain jls.net

            """
            mock_connect.side_effect = self._mock_manager
            mock_execute.side_effect = self._mock_manager
            self.dev2 = Device(
                host="2.2.2.2",
                user="test",
                password="password123",
                hostkey_verify=False,
            )
            self.dev2.open()
            self.assertEqual(self.dev2.connected, True)

    @patch("ncclient.manager.connect")
    @patch("jnpr.junos.Device.execute")
    def test_device_outbound(self, mock_connect, mock_execute):
        with patch("jnpr.junos.utils.fs.FS.cat") as mock_cat:
            mock_cat.return_value = """

    domain jls.net

            """
            mock_connect.side_effect = self._mock_manager
            mock_execute.side_effect = self._mock_manager
            self.dev2 = Device(sock_fd=6, user="test", password="password123")
            self.dev2.open()
            self.assertEqual(self.dev2.connected, True)

    @patch("jnpr.junos.Device.execute")
    def test_device_facts(self, mock_execute):
        with patch("jnpr.junos.utils.fs.FS.cat") as mock_cat:
            mock_execute.side_effect = self._mock_manager
            mock_cat.return_value = """

    domain jls.net

            """
            self.dev.facts_refresh()
            self.dev.facts._cache["current_re"] = ["re0"]
            assert self.dev.facts["version"] == facts["version"]

    @patch("jnpr.junos.Device.execute")
    @patch("jnpr.junos.factcache.warnings")
    def test_device_facts_error(self, mock_warnings, mock_execute):
        with patch("jnpr.junos.utils.fs.FS.cat") as mock_cat:
            mock_execute.side_effect = self._mock_manager
            mock_cat.side_effect = IOError("File cant be handled")
            self.dev.facts_refresh(warnings_on_failure=True)
            self.assertTrue(mock_warnings.warn.called)

    @patch("jnpr.junos.Device.execute")
    @patch("jnpr.junos.device.warnings")
    def test_device_facts_error_exception_on_error(self, mock_warnings, mock_execute):
        with patch("jnpr.junos.utils.fs.FS.cat") as mock_cat:
            mock_execute.side_effect = self._mock_manager
            mock_cat.side_effect = IOError("File cant be handled")
            self.assertRaises(
                IOError, self.dev.facts_refresh, exception_on_failure=True
            )

    @patch("jnpr.junos.Device.execute")
    @patch("jnpr.junos.device.warnings")
    def test_device_old_style_facts_error_exception_on_error(
        self, mock_warnings, mock_execute
    ):
        self.dev._fact_style = "old"
        with patch("jnpr.junos.utils.fs.FS.cat") as mock_cat:
            mock_execute.side_effect = self._mock_manager
            mock_cat.side_effect = IOError("File cant be handled")
            self.assertRaises(
                IOError, self.dev.facts_refresh, exception_on_failure=True
            )

    def test_device_facts_refresh_unknown_fact_style(self):
        self.dev._fact_style = "bad"
        with self.assertRaises(RuntimeError):
            self.dev.facts_refresh()

    def test_device_facts_refresh_old_fact_style_with_keys(self):
        self.dev._fact_style = "old"
        with self.assertRaises(RuntimeError):
            self.dev.facts_refresh(keys="domain")

    def test_device_hostname(self):
        self.assertEqual(self.dev.hostname, "1.1.1.1")

    def test_device_user(self):
        self.assertEqual(self.dev.user, "test")

    def test_device_get_password(self):
        self.assertEqual(self.dev.password, None)

    def test_device_set_password(self):
        self.dev.password = "secret"
        self.assertEqual(self.dev._auth_password, "secret")

    def test_device_get_timeout(self):
        self.assertEqual(self.dev.timeout, 30)

    def test_device_set_timeout(self):
        self.dev.timeout = 10
        self.assertEqual(self.dev.timeout, 10)

    def test_device_set_timeout_string(self):
        self.dev.timeout = "10"
        self.assertEqual(self.dev.timeout, 10)

    def test_device_set_timeout_invalid_string_value(self):
        with self.assertRaises(RuntimeError):
            self.dev.timeout = "foo"

    def test_device_set_timeout_invalid_type(self):
        with self.assertRaises(RuntimeError):
            self.dev.timeout = [1, 2, 3, 4]

    def test_device_manages(self):
        self.assertEqual(self.dev.manages, [], "By default manages will be empty list")

    @patch("ncclient.manager.connect")
    @patch("jnpr.junos.Device.execute")
    def test_device_open_normalize(self, mock_connect, mock_execute):
        mock_connect.side_effect = self._mock_manager
        self.dev2 = Device(host="2.2.2.2", user="test", password="password123")
        self.dev2.open(gather_facts=False, normalize=True)
        self.assertEqual(self.dev2.transform, self.dev2._norm_transform)

    def test_device_conn_None_transform(self):
        self.dev = Device(host="2.2.2.2", user="test", password="password123")
        with self.assertRaises(EzErrors.ConnectError):
            self.dev.transform

    def test_device_set_facts_exception(self):
        try:
            self.dev.facts = "test"
        except RuntimeError as ex:
            self.assertEqual(RuntimeError, type(ex))

    def test_device_ofacts_exception(self):
        with self.assertRaises(RuntimeError):
            ofacts = self.dev.ofacts

    def test_device_set_ofacts_exception(self):
        with self.assertRaises(RuntimeError):
            self.dev.ofacts = False

    @patch("jnpr.junos.Device.execute")
    def test_device_cli(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.assertEqual(self.dev.cli("show cli directory", warning=False).tag, "cli")

    @patch("jnpr.junos.device.json.loads")
    def test_device_rpc_json_ex(self, mock_json_loads):
        self.dev.facts = facts
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        ex = ValueError("Extra data ")
        ex.message = "Extra data "  # for py3 as we dont have message thr
        mock_json_loads.side_effect = [
            ex,
            self._mock_manager(
                etree.fromstring('<get-route-information format="json"/>')
            ),
        ]
        self.dev.rpc.get_route_information({"format": "json"})
        self.assertEqual(mock_json_loads.call_count, 2)

    @patch("jnpr.junos.Device.execute")
    def test_device_cli_to_rpc_string(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        data = self.dev.cli_to_rpc_string("show system uptime")
        self.assertEqual("rpc.get_system_uptime_information()", data)

    @patch("jnpr.junos.Device.execute")
    def test_device_cli_to_rpc_string_strip_pipes(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        data = self.dev.cli_to_rpc_string("show system uptime | match foo | count")
        self.assertEqual("rpc.get_system_uptime_information()", data)

    @patch("jnpr.junos.Device.execute")
    def test_device_cli_to_rpc_string_complex(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        data = self.dev.cli_to_rpc_string(
            "show interfaces ge-0/0/0.0 routing-instance all media"
        )
        self.assertEqual(
            "rpc.get_interface_information("
            "routing_instance='all', media=True, "
            "interface_name='ge-0/0/0.0')",
            data,
        )

    @patch("jnpr.junos.Device.execute")
    def test_device_cli_to_rpc_string_invalid(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        data = self.dev.cli_to_rpc_string("foo")
        self.assertEqual(None, data)

    @patch("jnpr.junos.Device.execute")
    def test_device_cli_format_json(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        data = self.dev.cli("show interface terse", warning=False, format="json")
        self.assertEqual(type(data), dict)
        self.assertEqual(
            data["interface-information"][0]["physical-interface"][0]["oper-status"][0][
                "data"
            ],
            "up",
        )

    @patch("jnpr.junos.Device.execute")
    def test_device_cli_conf_info(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.assertTrue("ge-0/0/0" in self.dev.cli("show configuration", warning=False))

    @patch("jnpr.junos.Device.execute")
    def test_device_cli_output(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.assertTrue("Alarm" in self.dev.cli("show system alarms", warning=False))

    @patch("jnpr.junos.Device.execute")
    @patch("jnpr.junos.device.warnings")
    def test_device_cli_output_warning(self, mock_warnings, mock_execute):
        mock_execute.side_effect = self._mock_manager
        data = self.dev.cli(
            "show interfaces ge-0/0/0.0 routing-instance " "all media", format="xml"
        )
        ip = data.findtext(
            'logical-interface[name="ge-0/0/0.0"]/'
            'address-family[address-family-name="inet"]/'
            "interface-address/ifa-local"
        )
        self.assertTrue("192.168.100.1" in ip)
        self.assertTrue(mock_warnings.warn.called)
        rpc_string = (
            "rpc.get_interface_information(routing_instance='all', "
            "media=True, interface_name='ge-0/0/0.0')"
        )
        self.assertIn(rpc_string, mock_warnings.warn.call_args[0][0])

    def test_device_cli_blank_output(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        self.assertEqual(
            "", self.dev.cli("show configuration interfaces", warning=False)
        )

    def test_device_cli_rpc_reply_with_message(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        self.assertEqual(
            "\nprotocol: operation-failed\nerror: device asdf not found\n",
            self.dev.cli("show interfaces terse asdf", warning=False),
        )

    @patch("jnpr.junos.Device.execute")
    def test_device_cli_rpc(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.assertEqual(
            self.dev.cli("show system uptime| display xml rpc", warning=False).tag,
            "get-system-uptime-information",
        )

    def test_device_cli_connection_exception(self):
        self.dev.connected = False
        self.assertRaises(EzErrors.ConnectClosedError, self.dev.cli, "foo")

    @patch("jnpr.junos.Device.execute")
    def test_device_cli_rpc_exception(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.assertRaises(EzErrors.RpcError, self.dev.cli, "foo")

    def test_device_cli_timeout_exception(self):
        self.dev._conn.rpc = MagicMock(side_effect=TimeoutExpiredError)
        self.assertRaises(EzErrors.RpcTimeoutError, self.dev.cli, "foo")

    @patch("jnpr.junos.device.warnings")
    def test_device_cli_unknown_exception(self, mock_warnings):
        class MyException(Exception):
            pass

        self.dev._conn.rpc = MagicMock(side_effect=MyException)
        self.assertRaises(MyException, self.dev.cli, "foo")

    @patch("jnpr.junos.Device.execute")
    def test_device_display_xml_rpc(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.assertEqual(
            self.dev.display_xml_rpc("show system uptime").tag,
            "get-system-uptime-information",
        )

    @patch("jnpr.junos.Device.execute")
    def test_device_display_xml_rpc_text(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.assertIn(
            "<get-system-uptime-information>",
            self.dev.display_xml_rpc("show system uptime", format="text"),
        )

    @patch("jnpr.junos.Device.execute")
    def test_device_display_xml_exception(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.assertEqual(
            self.dev.display_xml_rpc("show foo"),
            "invalid command: show foo| display xml rpc",
        )

    def test_device_execute(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        self.assertEqual(
            self.dev.execute("<get-system-core-dumps/>").tag, "directory-list"
        )

    def test_device_execute_topy(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        self.assertEqual(
            self.dev.execute("<get-system-core-dumps/>", to_py=self._do_nothing),
            "Nothing",
        )

    # This test is for the commented out rpc-error code
    #     def test_device_execute_exception(self):
    #         self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
    #         self.assertRaises(RpcError, self.dev.execute,
    #                           '<load-configuration-error/>')

    @patch("jnpr.junos.device.warnings")
    def test_device_execute_unknown_exception(self, mock_warnings):
        class MyException(Exception):
            pass

        self.dev._conn.rpc = MagicMock(side_effect=MyException)
        self.assertRaises(MyException, self.dev.execute, "<get-software-information/>")

    def test_device_execute_rpc_error(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        self.assertRaises(RpcError, self.dev.rpc.get_rpc_error)

    @unittest.skipIf(sys.platform == "win32", "will work for windows in coming days")
    def test_device_execute_permission_error(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        self.assertRaises(EzErrors.PermissionError, self.dev.rpc.get_permission_denied)

    def test_device_execute_index_error(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        self.assertTrue(self.dev.rpc.get_index_error())

    def test_device_execute_ValueError(self):
        self.assertRaises(ValueError, self.dev.execute, None)

    def test_device_execute_unopened(self):
        self.dev.connected = False
        self.assertRaises(EzErrors.ConnectClosedError, self.dev.execute, None)

    def test_device_execute_timeout(self):
        self.dev._conn.rpc = MagicMock(side_effect=TimeoutExpiredError)
        self.assertRaises(EzErrors.RpcTimeoutError, self.dev.rpc.get_rpc_timeout)

    def test_device_execute_closed(self):
        self.dev._conn.rpc = MagicMock(side_effect=NcErrors.TransportError)
        self.assertRaises(EzErrors.ConnectClosedError, self.dev.rpc.get_rpc_close)
        self.assertFalse(self.dev.connected)

    def test_device_rpcmeta(self):
        self.assertEqual(
            self.dev.rpc.get_software_information.__doc__, "get-software-information"
        )

    def test_device_probe_timeout_zero(self):
        with patch("jnpr.junos.device.socket"):
            self.assertFalse(self.dev.probe(0))

    def test_device_probe_timeout_gt_zero(self):
        with patch("jnpr.junos.device.socket"):
            self.assertTrue(
                self.dev.probe(1),
                "probe fn is not working for" " timeout greater than zero",
            )

    def test_device_probe_timeout_exception(self):
        with patch("jnpr.junos.device.socket") as mock_socket:
            with patch("jnpr.junos.device.time.sleep") as mock_time:
                mock_socket.socket.return_value.close.side_effect = RuntimeError
                mock_time.return_value = None
                self.assertFalse(self.dev.probe(0.01))

    def test_device_bind_varg(self):
        self.dev.bind()
        mock = MagicMock()
        mock.__name__ = "magic_mock"
        self.dev.bind(mock)
        self.assertEqual(self.dev.magic_mock.__name__, "magic_mock")

    def test_device_bind_kvarg(self):
        self.dev.bind()
        mock = MagicMock()
        mock.return_value = "Test"
        self.dev.bind(kw=mock)
        self.assertEqual(self.dev.kw, "Test")

    def test_device_bind_varg_exception(self):
        def varg():
            self.dev.bind()
            mock = MagicMock()
            mock.__name__ = "magic mock"
            # for *args
            self.dev.bind(mock)
            self.dev.bind(mock)

        self.assertRaises(ValueError, varg)

    def test_device_bind_kvarg_exception(self):
        def kve():
            self.dev.bind()
            mock = MagicMock()
            mock.__name__ = "magic mock"
            # for **kwargs
            self.dev.bind(kw=mock)
            self.dev.bind(kw=mock)

        self.assertRaises(ValueError, kve)

    def test_device_template(self):
        # Try to load the template relative to module base
        try:
            template = self.dev.Template("tests/unit/templates/config-example.xml")
        except:
            # Try to load the template relative to test base
            try:
                template = self.dev.Template("templates/config-example.xml")
            except:
                raise
        self.assertEqual(
            template.render({"host_name": "1", "domain_name": "2"}),
            "system {\n  host-name 1;\n  domain-name 2;\n}",
        )

    def test_device_close(self):
        def close_conn():
            self.dev.connected = False

        self.dev.close = MagicMock(name="close")
        self.dev.close.side_effect = close_conn
        self.dev.close()
        self.assertEqual(self.dev.connected, False)

    @patch("ncclient.manager.connect")
    def test_device_context_manager(self, mock_connect):
        mock_connect.side_effect = self._mock_manager
        try:
            with Device(
                host="3.3.3.3", user="gic", password="password123", gather_facts=False
            ) as dev:
                self.assertTrue(dev.connected)
                dev._conn = MagicMock(name="_conn")
                dev._conn.connected = True

                def close_conn():
                    dev.connected = False

                dev.close = MagicMock(name="close")
                dev.close.side_effect = close_conn
                raise RpcError
        except Exception as e:
            self.assertIsInstance(e, RpcError)
        self.assertFalse(dev.connected)

    def _read_file(self, fname):
        from ncclient.xml_ import NCElement

        fpath = os.path.join(os.path.dirname(__file__), "rpc-reply", fname)
        with open(fpath) as fp:
            foo = fp.read()

            if fname == "get-rpc-error.xml":
                # Raise ncclient exception for error
                raise RPCError(etree.XML(foo))
            elif fname == "get-permission-denied.xml":
                # Raise ncclient exception for error
                raise RPCError(etree.XML(foo))
            elif (
                fname == "get-index-error.xml"
                or fname == "get-system-core-dumps.xml"
                or fname == "load-configuration-error.xml"
                or fname == "show-configuration-interfaces.xml"
                or fname == "show-interfaces-terse-asdf.xml"
            ):
                rpc_reply = NCElement(
                    foo, self.dev._conn._device_handler.transform_reply()
                )
            elif fname == "show-configuration.xml" or fname == "show-system-alarms.xml":
                rpc_reply = NCElement(
                    foo, self.dev._conn._device_handler.transform_reply()
                )._NCElement__doc
            elif fname == "show-interface-terse.json":
                rpc_reply = json.loads(foo)
            elif fname == "get-route-information.json":
                rpc_reply = NCElement(
                    foo, self.dev._conn._device_handler.transform_reply()
                )
            else:
                rpc_reply = NCElement(
                    foo, self.dev._conn._device_handler.transform_reply()
                )._NCElement__doc[0]
        return rpc_reply

    def _mock_manager(self, *args, **kwargs):
        if kwargs and "normalize" not in kwargs:
            device_params = kwargs["device_params"]
            device_handler = make_device_handler(device_params)
            session = SSHSession(device_handler)
            return Manager(session, device_handler)
        elif args:
            if args[0].tag == "command":
                if args[0].text == "show cli directory":
                    return self._read_file("show-cli-directory.xml")
                if args[0].text == "show interface terse":
                    return self._read_file("show-interface-terse.json")
                elif args[0].text == "show configuration":
                    return self._read_file("show-configuration.xml")
                elif args[0].text == "show system alarms":
                    return self._read_file("show-system-alarms.xml")
                elif args[0].text == "show system uptime| display xml rpc":
                    return self._read_file("show-system-uptime-rpc.xml")
                elif args[0].text == "show configuration interfaces":
                    return self._read_file("show-configuration-interfaces.xml")
                elif args[0].text == "show interfaces terse asdf":
                    return self._read_file("show-interfaces-terse-asdf.xml")
                elif (
                    args[0].text == "show interfaces ge-0/0/0.0 "
                    "routing-instance all media"
                ):
                    return self._read_file("show-interfaces-routing-instance-media.xml")
                elif (
                    args[0].text == "show interfaces ge-0/0/0.0 "
                    "routing-instance all media| display "
                    "xml rpc"
                ):
                    return self._read_file(
                        "show-interfaces-routing-instance-media-rpc.xml"
                    )
                else:
                    raise RpcError

            else:
                if args[0].attrib.get("format") == "json":
                    return self._read_file(args[0].tag + ".json")
                return self._read_file(args[0].tag + ".xml")

    def _do_nothing(self, *args, **kwargs):
        return "Nothing"



import unittest
import nose2
import os

from ncclient.manager import Manager, make_device_handler
from ncclient.transport import SSHSession

from jnpr.junos import Device
from jnpr.junos.utils.fs import FS
from jnpr.junos.exception import RpcError

from unittest.mock import patch, MagicMock, call
from lxml import etree

__author__ = "Nitin Kumar, Rick Sherman"
__credits__ = "Jeremy Schulman"


class TestFS(unittest.TestCase):
    @patch("ncclient.manager.connect")
    def setUp(self, mock_connect):
        mock_connect.side_effect = self._mock_manager
        self.dev = Device(
            host="1.1.1.1", user="rick", password="password123", gather_facts=False
        )
        self.dev.open()
        self.fs = FS(self.dev)

    @patch("jnpr.junos.device.warnings")
    def test_cat_wrong_path_return_none(self, mock_warnings):
        path = "test/report"
        self.assertEqual(self.fs.cat(path), None)

    def test_cat(self):
        self.fs._dev.rpc.file_show = MagicMock(side_effect=self._mock_manager)
        path = "test/cat.txt"
        self.assertTrue("testing cat functionality" in self.fs.cat(path))
        self.fs._dev.rpc.file_show.assert_called_with(filename="test/cat.txt")

    def test_cwd(self):
        self.fs._dev.rpc.set_cli_working_directory = MagicMock(
            side_effect=self._mock_manager
        )
        folder = "change/directory"
        self.assertEqual("change/directory", self.fs.cwd(folder))
        self.fs._dev.rpc.set_cli_working_directory.assert_called_with(
            directory="change/directory"
        )

    @patch("jnpr.junos.Device.execute")
    def test_pwd(self, mock_execute):
        mock_execute.side_effect = MagicMock(side_effect=self._mock_manager)
        self.fs.pwd()
        self.assertEqual(self.fs.pwd(), "/cf/var/home/rick")

    @patch("jnpr.junos.device.warnings")
    def test_checksum_return_none(self, mock_warnings):
        path = "test/report"
        self.assertEqual(self.fs.checksum(path), None)

    def test_checksum_unknown_calc(self):
        path = "test/report"
        self.assertRaises(ValueError, self.fs.checksum, path=path, calc="abc")

    def test_checksum_return_rsp(self):
        self.fs.dev.rpc.get_sha256_checksum_information = MagicMock(
            side_effect=self._mock_manager
        )
        path = "test/checksum"
        self.assertEqual(self.fs.checksum(path, "sha256"), "xxxx")
        self.fs.dev.rpc.get_sha256_checksum_information.assert_called_with(
            path="test/checksum"
        )

    def test_stat_calling___decode_file(self):
        path = "test/stat/decode_file"
        self.fs.dev.rpc.file_list = MagicMock(side_effect=self._mock_manager)
        self.assertEqual(
            self.fs.stat(path),
            {
                "owner": "pqr",
                "path": "/var/abc.sh",
                "permissions": 755,
                "permissions_text": "-rwxr-xr-x",
                "size": 2,
                "ts_date": "Mar 13 06:54",
                "ts_epoc": "1394693680",
                "type": "file",
            },
        )

    def test_stat_calling___decode_dir(self):
        path = "test/stat/decode_dir"
        self.fs.dev.rpc.file_list = MagicMock(side_effect=self._mock_manager)
        self.assertEqual(
            self.fs.stat(path),
            {"path": "/var", "type": "dir", "file_count": 1, "size": 2},
        )

    def test_stat_return_none(self):
        path = "test/abc"
        self.fs.dev.rpc.file_list = MagicMock()
        self.fs.dev.rpc.file_list.find.return_value = "output"
        self.assertEqual(self.fs.stat(path), None)

    def test_ls_calling___decode_file(self):
        path = "test/stat/decode_file"
        self.fs.dev.rpc.file_list = MagicMock(side_effect=self._mock_manager)
        self.assertEqual(
            self.fs.ls(path),
            {
                "owner": "pqr",
                "path": "/var/abc.sh",
                "permissions": 755,
                "permissions_text": "-rwxr-xr-x",
                "size": 2,
                "ts_date": "Mar 13 06:54",
                "ts_epoc": "1394693680",
                "type": "file",
            },
        )

    def test_ls_calling___decode_dir(self):
        path = "test/stat/decode_dir"
        self.fs.dev.rpc.file_list = MagicMock(side_effect=self._mock_manager)
        self.assertEqual(
            self.fs.ls(path),
            {
                "files": {
                    "abc": {
                        "permissions_text": "drwxr-xr-x",
                        "ts_date": "Feb 17 15:30",
                        "ts_epoc": "1392651039",
                        "owner": "root",
                        "path": "abc",
                        "size": 2,
                        "type": "dir",
                        "permissions": 555,
                    }
                },
                "path": "/var",
                "type": "dir",
                "file_count": 1,
                "size": 2,
            },
        )

    def test_ls_return_none(self):
        path = "test/abc"
        self.fs.dev.rpc.file_list = MagicMock()
        self.fs.dev.rpc.file_list.find.return_value = "output"
        self.assertEqual(self.fs.ls(path), None)

    @patch("jnpr.junos.utils.fs.FS._decode_file")
    def test_ls_link_path_false(self, mock_decode_file):
        mock_decode_file.get.return_value = False
        path = "test/stat/decode_file"
        self.fs.dev.rpc.file_list = MagicMock(side_effect=self._mock_manager)
        self.fs.ls(path, followlink=False)
        mock_decode_file.assert_has_calls([call().get("link")])

    def test_ls_brief_true(self):
        path = "test/stat/decode_dir"
        self.fs.dev.rpc.file_list = MagicMock(side_effect=self._mock_manager)
        self.assertEqual(
            self.fs.ls(path, brief=True),
            {
                "files": ["abc"],
                "path": "/var",
                "type": "dir",
                "file_count": 1,
                "size": 2,
            },
        )

    def test_ls_calling___decode_dir_type_symbolic_link(self):
        path = "test/stat/decode_symbolic_link"
        self.fs.dev.rpc.file_list = MagicMock(side_effect=self._mock_manager)
        self.assertEqual(
            self.fs.ls(path),
            {
                "files": {
                    "abc": {
                        "permissions_text": "drwxr-xr-x",
                        "ts_date": "Feb 17 15:30",
                        "link": "symlink test",
                        "ts_epoc": "1392651039",
                        "owner": "root",
                        "path": "abc",
                        "size": 2,
                        "type": "link",
                        "permissions": 555,
                    }
                },
                "path": "/var",
                "type": "dir",
                "file_count": 1,
                "size": 2,
            },
        )

    def test_rm_return_true(self):
        self.fs.dev.rpc.file_delete = MagicMock(return_value=True)
        path = "test/abc"
        self.assertTrue(self.fs.rm(path))
        self.fs.dev.rpc.file_delete.assert_called_once_with(path="test/abc")

    def test_rm_return_false(self):
        path = "test/abc"
        self.fs.dev.rpc.file_delete = MagicMock(return_value=False)
        self.assertFalse(self.fs.rm(path))
        self.fs.dev.rpc.file_delete.assert_called_once_with(path="test/abc")

    def test_copy_return_true(self):
        self.fs.dev.rpc.file_copy = MagicMock()
        initial = "test/abc"
        final = "test/xyz"
        self.assertTrue(self.fs.cp(initial, final))
        self.fs.dev.rpc.file_copy.assert_called_once_with(
            source="test/abc", destination="test/xyz"
        )

    def test_copy_routing_instance_return_true(self):
        self.fs.dev.rpc.file_copy = MagicMock()
        initial = "test/abc"
        final = "test/xyz"
        ri_name = "test_ri"
        self.assertTrue(self.fs.cp(initial, final, routing_instance=ri_name))
        self.fs.dev.rpc.file_copy.assert_called_once_with(
            source="test/abc", destination="test/xyz", routing_instance="test_ri"
        )

    def test_copy_source_address_return_true(self):
        self.fs.dev.rpc.file_copy = MagicMock()
        initial = "test/abc"
        final = "test/xyz"
        s_add = "0.0.0.0"
        self.assertTrue(self.fs.cp(initial, final, source_address=s_add))
        self.fs.dev.rpc.file_copy.assert_called_once_with(
            source="test/abc", destination="test/xyz", source_address="0.0.0.0"
        )

    def test_copy_return_false(self):
        initial = "test/abc"
        final = "test/xyz"
        self.fs.dev.rpc.file_copy = MagicMock(side_effect=Exception)
        self.assertFalse(self.fs.cp(initial, final))
        self.fs.dev.rpc.file_copy.assert_called_once_with(
            source="test/abc", destination="test/xyz"
        )

    def test_move_return_true(self):
        self.fs.dev.rpc.file_rename = MagicMock(return_value=True)
        initial = "test/abc"
        final = "test/xyz"
        self.assertTrue(self.fs.mv(initial, final))
        self.fs.dev.rpc.file_rename.assert_called_once_with(
            source="test/abc", destination="test/xyz"
        )

    def test_move_return_false(self):
        initial = "test/abc"
        final = "test/xyz"
        self.fs.dev.rpc.file_rename = MagicMock(return_value=False)
        self.assertFalse(self.fs.mv(initial, final))
        self.fs.dev.rpc.file_rename.assert_called_once_with(
            source="test/abc", destination="test/xyz"
        )

    def test_tgz_return_true(self):
        src = "test/tgz.txt"
        dst = "test/xyz"
        self.fs.dev.rpc.file_archive = MagicMock(return_value=True)
        self.assertTrue(self.fs.tgz(src, dst))
        self.fs.dev.rpc.file_archive.assert_called_once_with(
            source="test/tgz.txt", destination="test/xyz", compress=True
        )

    @patch("jnpr.junos.Device.execute")
    def test_tgz_return_error(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        src = "test/tgz.txt"
        dst = "test/xyz"
        self.assertTrue("testing tgz" in self.fs.tgz(src, dst))

    @patch("jnpr.junos.utils.fs.StartShell")
    def test_rmdir(self, mock_StartShell):
        path = "test/rmdir"
        print(self.fs.rmdir(path))
        calls = [
            call().__enter__(),
            call().__enter__().run("rmdir test/rmdir"),
            call().__exit__(None, None, None),
        ]
        mock_StartShell.assert_has_calls(calls)

    @patch("jnpr.junos.utils.fs.StartShell")
    def test_mkdir(self, mock_StartShell):
        path = "test/mkdir"
        print(self.fs.mkdir(path))
        calls = [
            call().__enter__(),
            call().__enter__().run("mkdir -p test/mkdir"),
            call().__exit__(None, None, None),
        ]
        mock_StartShell.assert_has_calls(calls)

    @patch("jnpr.junos.utils.fs.StartShell")
    def test_symlink(self, mock_StartShell):
        src = "test/tgz.txt"
        dst = "test/xyz"
        print(self.fs.symlink(src, dst))
        calls = [
            call().__enter__(),
            call().__enter__().run("ln -sf test/tgz.txt test/xyz"),
            call().__exit__(None, None, None),
        ]
        mock_StartShell.assert_has_calls(calls)

    @patch("jnpr.junos.Device.execute")
    def test_storage_usage(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.assertEqual(
            self.fs.storage_usage(),
            {
                "/dev/abc": {
                    "avail_block": 234234,
                    "used_blocks": 2346455,
                    "used_pct": "1",
                    "mount": "/",
                    "total_blocks": 567431,
                    "avail": "2F",
                    "used": "481M",
                    "total": "4F",
                }
            },
        )

    @patch("jnpr.junos.Device.execute")
    def test_storage_usage_linux(self, mock_execute):
        mock_execute.side_effect = self._mock_manager_linux
        self.assertEqual(
            self.fs.storage_usage(),
            {
                "re0": {
                    "/dev/sda6": {
                        "avail": "916M",
                        "avail_block": 1874712,
                        "mount": "/data/config",
                        "total": "984M",
                        "total_blocks": 2015024,
                        "used": "1.4M",
                        "used_blocks": 2688,
                        "used_pct": "1",
                    }
                }
            },
        )

    @patch("jnpr.junos.Device.execute")
    def test_directory_usage(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.assertEqual(
            self.fs.directory_usage(path="/var/tmp", depth=1),
            {
                "/var/tmp": {"blocks": 456076, "bytes": 233510912, "size": "223M"},
                "/var/tmp/gres-tp": {"blocks": 68, "bytes": 34816, "size": "34K"},
                "/var/tmp/install": {"blocks": 4, "bytes": 2048, "size": "2.0K"},
                "/var/tmp/pics": {"blocks": 4, "bytes": 2048, "size": "2.0K"},
                "/var/tmp/rtsdb": {"blocks": 4, "bytes": 2048, "size": "2.0K"},
                "/var/tmp/sec-download": {"blocks": 8, "bytes": 4096, "size": "4.0K"},
                "/var/tmp/vi.recover": {"blocks": 4, "bytes": 2048, "size": "2.0K"},
            },
        )

    @patch("jnpr.junos.Device.execute")
    def test_directory_usage_error(self, mock_execute):
        mock_execute.return_value = etree.fromstring(
            """
            <directory-usage-information>
                <directory>
                    <used-space used-blocks="456076">
                        223M
                    </used-space>
                </directory>
            </directory-usage-information>"""
        )
        self.assertRaises(RpcError, self.fs.directory_usage, path="/var/tmp", depth=1)

    @patch("jnpr.junos.Device.execute")
    def test_directory_usage_no_directory(self, mock_execute):
        mock_execute.side_effect = self._mock_manager_error1
        self.assertRaises(RpcError, self.fs.directory_usage, path="/var/tmp", depth="1")

    @patch("jnpr.junos.Device.execute")
    def test_directory_usage_no_dir_name(self, mock_execute):
        mock_execute.side_effect = self._mock_manager_error2
        self.assertRaises(RpcError, self.fs.directory_usage, path="/var/tmp", depth="1")

    @patch("jnpr.junos.Device.execute")
    def test_storage_cleanup(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.assertEqual(
            self.fs.storage_cleanup(),
            {"/var/abc.txt": {"ts_date": "Apr 25 10:38", "size": 11}},
        )

    @patch("jnpr.junos.Device.execute")
    def test_storage_cleanup_check(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.assertEqual(
            self.fs.storage_cleanup_check(),
            {"/var/abc.txt": {"ts_date": "Apr 25 10:38", "size": 11}},
        )

    def _read_file(self, fname):
        from ncclient.xml_ import NCElement

        fpath = os.path.join(os.path.dirname(__file__), "rpc-reply", fname)
        foo = open(fpath).read()

        if (
            fname == "get-rpc-error.xml"
            or fname == "get-index-error.xml"
            or fname == "get-system-core-dumps.xml"
        ):
            rpc_reply = NCElement(foo, self.dev._conn._device_handler.transform_reply())
        elif (
            fname == "show-configuration.xml"
            or fname == "show-system-alarms.xml"
            or fname == "set-cli-working-directory.xml"
        ):
            rpc_reply = NCElement(
                foo, self.dev._conn._device_handler.transform_reply()
            )._NCElement__doc
        else:
            rpc_reply = NCElement(
                foo, self.dev._conn._device_handler.transform_reply()
            )._NCElement__doc[0]
        return rpc_reply

    def _mock_manager(self, *args, **kwargs):
        if kwargs:
            # if 'path' in kwargs and 'detail' in kwargs:
            #     return self._read_file('dir_list_detail.xml')

            if "path" in kwargs:
                if kwargs["path"] == "test/stat/decode_dir":
                    return self._read_file("file-list_dir.xml")
                elif kwargs["path"] == "test/stat/decode_file":
                    return self._read_file("file-list_file.xml")
                elif kwargs["path"] == "test/checksum":
                    return self._read_file("checksum.xml")
                elif kwargs["path"] == "test/stat/decode_symbolic_link":
                    return self._read_file("file-list_symlink.xml")
            if "directory" in kwargs:
                if kwargs["directory"] == "change/directory":
                    return self._read_file("set-cli-working-directory.xml")
            if "filename" in kwargs:
                if kwargs["filename"] == "test/cat.txt":
                    return self._read_file("file-show.xml")
            device_params = kwargs["device_params"]
            device_handler = make_device_handler(device_params)
            session = SSHSession(device_handler)
            return Manager(session, device_handler)

        elif args:
            if args[0].tag == "command":
                if args[0].text == "show cli directory":
                    return self._read_file("show-cli-directory.xml")
            elif args[0].tag == "get-system-storage":
                return self._read_file("get-system-storage.xml")
            elif args[0].tag == "get-directory-usage-information":
                return self._read_file("get-directory-usage-information.xml")
            elif args[0].tag == "request-system-storage-cleanup":
                return self._read_file("request-system-storage-cleanup.xml")
            elif args[0].tag == "file-archive":
                return self._read_file("file-archive.xml")

    def _mock_manager_error1(self, *args, **kwargs):
        if args:
            if args[0].tag == "get-directory-usage-information":
                return self._read_file("get-directory-usage-information_error1.xml")

    def _mock_manager_error2(self, *args, **kwargs):
        if args:
            if args[0].tag == "get-directory-usage-information":
                return self._read_file("get-directory-usage-information_error2.xml")

    def _mock_manager_linux(self, *args, **kwargs):
        if args:
            if args[0].tag == "get-system-storage":
                return self._read_file("get-storage-usage-linux.xml")



__author__ = "Stacy Smith"
__credits__ = "Jeremy Schulman, Nitin Kumar"

try:
    import unittest2 as unittest
except ImportError:
    import unittest
import nose2
import importlib
import sys

import jnpr.junos.facts


class TestFactInitialization(unittest.TestCase):
    def test_duplicate_facts(self):
        module = importlib.import_module("tests.unit.facts.dupe_foo1")
        sys.modules["jnpr.junos.facts.dupe_foo1"] = module
        module = importlib.import_module("tests.unit.facts.dupe_foo2")
        sys.modules["jnpr.junos.facts.dupe_foo2"] = module
        with self.assertRaises(RuntimeError):
            jnpr.junos.facts._build_fact_callbacks_and_doc_strings()



__author__ = "rsherman, vnitinv"

import unittest

from jnpr.junos import Device


class TestDeviceSsh(unittest.TestCase):
    def tearDown(self):
        self.dev.close()

    def test_device_open_key_pass(self):
        self.dev = Device(
            host="xxxx",
            user="jenkins",
            ssh_private_key_file="/var/lib/jenkins/.ssh/passkey",
            passwd="password",
        )
        self.dev.open()
        self.assertEqual(self.dev.connected, True)



__author__ = "Nitin Kumar, Rick Sherman"
__credits__ = "Jeremy Schulman"

try:
    import unittest2 as unittest
except ImportError:
    import unittest
import nose2
from unittest.mock import patch, MagicMock
import os

from jnpr.junos import Device
from jnpr.junos.ofacts.swver import facts_software_version as software_version
from jnpr.junos.ofacts.swver import _get_swver
from ncclient.manager import Manager, make_device_handler
from ncclient.transport import SSHSession
from jnpr.junos.exception import RpcError


class TestSwver(unittest.TestCase):
    @patch("ncclient.manager.connect")
    @patch("jnpr.junos.device.warnings")
    def setUp(self, mock_warnings, mock_connect):
        mock_connect.side_effect = self._mock_manager
        self.dev = Device(
            host="1.1.1.1",
            user="rick",
            password="password123",
            gather_facts=False,
            fact_style="old",
        )
        self.dev.open()
        self.facts = {}
        self.facts["vc_capable"] = False

    def test_get_swver_vc(self):
        self.dev.rpc.cli = MagicMock()
        self.facts["vc_capable"] = True
        _get_swver(self.dev, self.facts)
        self.dev.rpc.cli.assert_called_with("show version all-members", format="xml")

    def test_get_swver_vc_capable_standalone(self):
        def raise_ex(*args):
            if args[0] == "show version all-members":
                raise RpcError()

        self.dev.rpc.cli = MagicMock(
            side_effect=lambda *args, **kwargs: raise_ex(*args)
        )
        self.facts["vc_capable"] = True
        _get_swver(self.dev, self.facts)
        self.dev.rpc.cli.assert_called_with(
            "show version invoke-on all-routing-engines", format="xml"
        )

    @patch("jnpr.junos.Device.execute")
    def test_swver(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.facts["master"] = "RE0"
        software_version(self.dev, self.facts)
        self.assertEqual(self.facts["version"], "12.3R6.6")

    @patch("jnpr.junos.Device.execute")
    def test_swver_f_master_list(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.facts["master"] = ["RE0", "RE1"]
        software_version(self.dev, self.facts)
        self.assertEqual(self.facts["version"], "12.3R6.6")

    @patch("jnpr.junos.Device.execute")
    def test_swver_hostname_none(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.facts["master"] = "RE5"
        self.facts["version_RE5"] = "15.3R6.6"
        software_version(self.dev, self.facts)
        self.assertEqual(self.facts["version"], "15.3R6.6")

    @patch("jnpr.junos.Device.execute")
    def test_swver_txp_master_list(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.facts["master"] = ["RE0", "RE0", "RE1", "RE2", "RE3"]
        self.facts["version_RE0-RE0"] = "14.2R4"
        software_version(self.dev, self.facts)
        self.assertEqual(self.facts["version"], "14.2R4")

    # --> JLS, there should always be a facts['master'] assigned.
    # @patch('jnpr.junos.Device.execute')
    # def test_swver_master_none(self, mock_execute):
    #     mock_execute.side_effect = self._mock_manager
    #     self.facts['master'] = None
    #     software_version(self.dev, self.facts)
    #     self.assertEqual(self.facts['version'], '12.3R6.6')

    @patch("jnpr.junos.Device.execute")
    @patch("jnpr.junos.facts.get_software_information.re.findall")
    def test_swver_exception_handling(self, mock_re_findall, mock_execute):
        mock_execute.side_effect = self._mock_manager
        mock_re_findall.side_effect = IndexError
        self.facts["master"] = "RE0"
        software_version(self.dev, self.facts)
        self.assertEqual(self.facts["version"], "0.0I0.0")

    def _read_file(self, fname):
        from ncclient.xml_ import NCElement

        fpath = os.path.join(os.path.dirname(__file__), "rpc-reply", fname)
        foo = open(fpath).read()

        rpc_reply = NCElement(
            foo, self.dev._conn._device_handler.transform_reply()
        )._NCElement__doc[0]
        return rpc_reply

    def _mock_manager(self, *args, **kwargs):
        if kwargs:
            device_params = kwargs["device_params"]
            device_handler = make_device_handler(device_params)
            session = SSHSession(device_handler)
            return Manager(session, device_handler)

        if args:
            if "version_RE0-RE0" in self.facts:
                return self._read_file(args[0].tag + "_RE0-RE0.xml")
            return self._read_file(args[0].tag + ".xml")



__author__ = "Stacy Smith"
__credits__ = "Jeremy Schulman, Nitin Kumar"

import unittest
import nose2
from unittest.mock import patch, MagicMock
import os

from jnpr.junos import Device

from ncclient.manager import Manager, make_device_handler
from ncclient.transport import SSHSession


class TestFileList(unittest.TestCase):
    @patch("ncclient.manager.connect")
    def setUp(self, mock_connect):
        mock_connect.side_effect = self._mock_manager_setup
        self.dev = Device(
            host="1.1.1.1", user="rick", password="password123", gather_facts=False
        )
        self.dev.open()

    @patch("jnpr.junos.Device.execute")
    def test_home_fact(self, mock_execute):
        mock_execute.side_effect = self._mock_manager_file_list
        self.assertEqual(self.dev.facts["HOME"], "/var/home/user")

    @patch("jnpr.junos.Device.execute")
    def test_home_fact_multi_dir(self, mock_execute):
        mock_execute.side_effect = self._mock_manager_file_list2
        self.assertEqual(self.dev.facts["HOME"], "/var/home/user")

    def _read_file(self, fname):
        from ncclient.xml_ import NCElement

        fpath = os.path.join(os.path.dirname(__file__), "rpc-reply", fname)
        foo = open(fpath).read()

        rpc_reply = NCElement(
            foo, self.dev._conn._device_handler.transform_reply()
        )._NCElement__doc[0]
        return rpc_reply

    def _mock_manager_setup(self, *args, **kwargs):
        if kwargs:
            device_params = kwargs["device_params"]
            device_handler = make_device_handler(device_params)
            session = SSHSession(device_handler)
            return Manager(session, device_handler)

    def _mock_manager_file_list(self, *args, **kwargs):
        if args:
            return self._read_file("file_list_" + args[0].tag + ".xml")

    def _mock_manager_file_list2(self, *args, **kwargs):
        if args:
            return self._read_file("file_list2_" + args[0].tag + ".xml")



__author__ = "Stacy Smith"
__credits__ = "Jeremy Schulman, Nitin Kumar"

import unittest
import nose2
from unittest.mock import patch, MagicMock
import os
import sys
from lxml import etree

from jnpr.junos import Device
from jnpr.junos.exception import RpcError

from ncclient.manager import Manager, make_device_handler
from ncclient.transport import SSHSession


class TestGetVirtualChassisInformation(unittest.TestCase):
    @patch("ncclient.manager.connect")
    def setUp(self, mock_connect):
        mock_connect.side_effect = self._mock_manager_setup
        self.dev = Device(
            host="1.1.1.1", user="rick", password="password123", gather_facts=False
        )
        self.dev.open()

    @patch("jnpr.junos.Device.execute")
    def test_vc(self, mock_execute):
        mock_execute.side_effect = self._mock_manager_vc
        self.assertEqual(self.dev.facts["vc_capable"], True)
        self.assertEqual(self.dev.facts["vc_mode"], "Enabled")
        self.assertEqual(self.dev.facts["vc_fabric"], None)
        self.assertEqual(self.dev.facts["vc_master"], "0")

    @patch("jnpr.junos.Device.execute")
    def test_vc_dual_master(self, mock_execute):
        mock_execute.side_effect = self._mock_manager_vc_dual_master
        self.assertEqual(self.dev.facts["vc_capable"], None)
        self.assertEqual(self.dev.facts["vc_mode"], None)
        self.assertEqual(self.dev.facts["vc_fabric"], None)
        self.assertEqual(self.dev.facts["vc_master"], None)

    @patch("jnpr.junos.Device.execute")
    def test_vc_error(self, mock_execute):
        mock_execute.side_effect = self._mock_manager_vc_error
        self.assertEqual(self.dev.facts["vc_capable"], False)
        self.assertEqual(self.dev.facts["vc_mode"], None)
        self.assertEqual(self.dev.facts["vc_fabric"], None)
        self.assertEqual(self.dev.facts["vc_master"], None)

    @patch("jnpr.junos.Device.execute")
    def test_vc_empty(self, mock_execute):
        mock_execute.side_effect = self._mock_manager_vc_empty
        self.assertEqual(self.dev.facts["vc_capable"], False)
        self.assertEqual(self.dev.facts["vc_mode"], None)
        self.assertEqual(self.dev.facts["vc_fabric"], None)
        self.assertEqual(self.dev.facts["vc_master"], None)

    @patch("jnpr.junos.Device.execute")
    @unittest.skipIf(sys.platform == "win32", "will work for windows in coming days")
    def test_vc_mmvcf(self, mock_execute):
        mock_execute.side_effect = self._mock_manager_vc_mmvcf
        self.assertEqual(self.dev.facts["vc_capable"], True)
        self.assertEqual(self.dev.facts["vc_mode"], "Mixed")
        self.assertEqual(self.dev.facts["vc_fabric"], True)
        self.assertEqual(self.dev.facts["vc_master"], "0")

    @patch("jnpr.junos.Device.execute")
    @unittest.skipIf(sys.platform == "win32", "will work for windows in coming days")
    def test_vc_mmvc(self, mock_execute):
        mock_execute.side_effect = self._mock_manager_vc_mmvc
        self.assertEqual(self.dev.facts["vc_capable"], True)
        self.assertEqual(self.dev.facts["vc_mode"], "Mixed")
        self.assertEqual(self.dev.facts["vc_fabric"], False)
        self.assertEqual(self.dev.facts["vc_master"], "0")

    def _read_file(self, fname):
        from ncclient.xml_ import NCElement

        fpath = os.path.join(os.path.dirname(__file__), "rpc-reply", fname)
        foo = open(fpath).read()

        rpc_reply = NCElement(
            foo, self.dev._conn._device_handler.transform_reply()
        )._NCElement__doc[0]
        return rpc_reply

    def _mock_manager_setup(self, *args, **kwargs):
        if kwargs:
            device_params = kwargs["device_params"]
            device_handler = make_device_handler(device_params)
            session = SSHSession(device_handler)
            return Manager(session, device_handler)

    def _mock_manager_vc(self, *args, **kwargs):
        if args:
            return self._read_file("vc_" + args[0].tag + ".xml")

    def _mock_manager_vc_dual_master(self, *args, **kwargs):
        if args:
            return self._read_file("vc_dual_master_" + args[0].tag + ".xml")

    def _mock_manager_vc_mmvcf(self, *args, **kwargs):
        if args:
            return self._read_file("vc_mmvcf_" + args[0].tag + ".xml")

    def _mock_manager_vc_mmvc(self, *args, **kwargs):
        if args:
            return self._read_file("vc_mmvc_" + args[0].tag + ".xml")

    def _mock_manager_vc_error(self, *args, **kwargs):
        if args:
            raise RpcError

    def _mock_manager_vc_empty(self, *args, **kwargs):
        if args:
            return True



import unittest
import os
import re
import nose2

from jnpr.junos.device import Device
from jnpr.junos.rpcmeta import _RpcMetaExec
from jnpr.junos.facts.swver import version_info
from ncclient.manager import Manager, make_device_handler
from ncclient.transport import SSHSession
from jnpr.junos.exception import JSONLoadError

from unittest.mock import patch, MagicMock, call
from lxml import etree

__author__ = "Nitin Kumar, Rick Sherman"
__credits__ = "Jeremy Schulman"


class Test_RpcMetaExec(unittest.TestCase):
    @patch("ncclient.manager.connect")
    def setUp(self, mock_connect):
        mock_connect.side_effect = self._mock_manager
        self.dev = Device(
            host="1.1.1.1", user="rick", password="password123", gather_facts=False
        )
        self.dev.open()
        self.rpc = _RpcMetaExec(self.dev)

    def test_rpcmeta_constructor(self):
        self.assertTrue(isinstance(self.rpc._junos, Device))

    @patch("jnpr.junos.device.Device.execute")
    def test_rpcmeta_load_config(self, mock_execute_fn):
        root = etree.XML("<root><a>test</a></root>")
        self.rpc.load_config(root)
        self.assertEqual(mock_execute_fn.call_args[0][0].tag, "load-configuration")

    @patch("jnpr.junos.device.Device.execute")
    def test_rpcmeta_load_config_with_configuration_tag(self, mock_execute_fn):
        root = etree.XML("<configuration><root><a>test</a></root></configuration>")
        self.rpc.load_config(root)
        self.assertEqual(mock_execute_fn.call_args[0][0].tag, "load-configuration")

    @patch("jnpr.junos.device.Device.execute")
    def test_rpcmeta_load_config_option_action(self, mock_execute_fn):
        set_commands = """
            set system host-name test_rpc
            set system domain-name test.juniper.net
        """
        self.rpc.load_config(set_commands, action="set")
        self.assertEqual(mock_execute_fn.call_args[0][0].get("action"), "set")

    @patch("jnpr.junos.device.Device.execute")
    def test_rpcmeta_option_format(self, mock_execute_fn):
        set_commands = """
            set system host-name test_rpc
            set system domain-name test.juniper.net
        """
        self.rpc.load_config(set_commands, format="text")
        self.assertEqual(mock_execute_fn.call_args[0][0].get("format"), "text")

    @patch("jnpr.junos.device.Device.execute")
    def test_rpcmeta_option_format_json(self, mock_execute_fn):
        json_commands = """
            {
                "configuration" : {
                    "system" : {
                        "services" : {
                            "telnet" : [null]
                        }
                    }
                }
            }
        """
        self.rpc.load_config(json_commands, format="json")
        self.assertEqual(mock_execute_fn.call_args[0][0].get("format"), "json")

    @patch("jnpr.junos.device.Device.execute")
    def test_rpcmeta_exec_rpc_vargs(self, mock_execute_fn):
        self.rpc.system_users_information(dict(format="text"))
        self.assertEqual(mock_execute_fn.call_args[0][0].get("format"), "text")

    @patch("jnpr.junos.device.Device.execute")
    def test_rpcmeta_exec_rpc_kvargs_bool_true(self, mock_execute_fn):
        self.rpc.system_users_information(test=True)
        self.assertEqual(mock_execute_fn.call_args[0][0][0].tag, "test")
        self.assertEqual(mock_execute_fn.call_args[0][0][0].text, None)

    @patch("jnpr.junos.device.Device.execute")
    def test_rpcmeta_exec_rpc_kvargs_bool_False(self, mock_execute_fn):
        self.rpc.system_users_information(test=False)
        self.assertEqual(mock_execute_fn.call_args[0][0].find("test"), None)

    @patch("jnpr.junos.device.Device.execute")
    def test_rpcmeta_exec_rpc_kvargs_tuple(self, mock_execute_fn):
        self.rpc.system_users_information(set_data=("test", "foo"))
        self.assertEqual(mock_execute_fn.call_args[0][0][0].text, "test")
        self.assertEqual(mock_execute_fn.call_args[0][0][1].text, "foo")

    @patch("jnpr.junos.device.Device.execute")
    def test_rpcmeta_exec_rpc_kvargs_dict(self, mock_execute_fn):
        self.assertRaises(
            TypeError, self.rpc.system_users_information, dict_data={"test": "foo"}
        )

    @patch("jnpr.junos.device.Device.execute")
    def test_rpcmeta_exec_rpc_kvargs_list_with_dict(self, mock_execute_fn):
        self.assertRaises(
            TypeError,
            self.rpc.system_users_information,
            list_with_dict_data=[True, {"test": "foo"}],
        )

    @patch("jnpr.junos.device.Device.execute")
    def test_rpcmeta_exec_rpc_normalize(self, mock_execute_fn):
        self.rpc.any_ole_rpc(normalize=True)
        self.assertEqual(mock_execute_fn.call_args[1], {"normalize": True})

    @patch("jnpr.junos.device.Device.execute")
    def test_rpcmeta_get_config(self, mock_execute_fn):
        root = etree.XML("<root><a>test</a></root>")
        self.rpc.get_config(root)
        self.assertEqual(mock_execute_fn.call_args[0][0].tag, "get-configuration")

    def test_rpcmeta_exec_rpc_format_json_14_2(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        self.dev.facts._cache["version_info"] = version_info("14.2X46-D15.3")
        op = self.rpc.get_system_users_information(dict(format="json"))
        self.assertEqual(
            op["system-users-information"][0]["uptime-information"][0]["date-time"][0][
                "data"
            ],
            "4:43AM",
        )

    def test_rpcmeta_exec_rpc_format_json_gt_14_2(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        self.dev.facts._cache["version_info"] = version_info("15.1X46-D15.3")
        op = self.rpc.get_system_users_information(dict(format="json"))
        self.assertEqual(
            op["system-users-information"][0]["uptime-information"][0]["date-time"][0][
                "data"
            ],
            "4:43AM",
        )

    @patch("jnpr.junos.device.warnings")
    def test_rpcmeta_exec_rpc_format_json_lt_14_2(self, mock_warn):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        self.dev.facts._cache["version_info"] = version_info("13.1X46-D15.3")
        self.rpc.get_system_users_information(dict(format="json"))
        mock_warn.assert_has_calls(
            [call.warn("Native JSON support is only from 14.2 onwards", RuntimeWarning)]
        )

    def test_get_rpc(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        resp = self.dev.rpc.get(filter_select="bgp")
        self.assertEqual(resp.tag, "data")

    def test_get_config_filter_xml_string_xml(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        resp = self.dev.rpc.get_config(filter_xml="<system><services/></system>")
        self.assertEqual(resp.tag, "configuration")

    def test_get_config_filter_xml_string(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        resp = self.dev.rpc.get_config(filter_xml="system/services")
        self.assertEqual(resp.tag, "configuration")

    def test_get_config_filter_xml_model(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        resp = self.dev.rpc.get_config("bgp/neighbors", model="openconfig")
        self.assertEqual(resp.tag, "bgp")

    def test_get_rpc_ignore_warning_bool(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        resp = self.dev.rpc.get(ignore_warning=True)
        self.assertEqual(resp.tag, "data")

    def test_get_rpc_ignore_warning_str(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        resp = self.dev.rpc.get(ignore_warning="vrrp subsystem not running")
        self.assertEqual(resp.tag, "data")

    def test_get_rpc_ignore_warning_list(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        resp = self.dev.rpc.get(
            ignore_warning=["vrrp subsystem not running", "statement not found"]
        )
        self.assertEqual(resp.tag, "data")

    # below test need to be fixed for Python 3.x
    """
    def test_get_config_remove_ns(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        resp = self.dev.rpc.get_config('bgp/neighbors', model='openconfig',
                                       remove_ns=False)
        self.assertEqual(resp.tag, '{http://openconfig.net/yang/bgp}bgp')
    """
    #

    def test_model_true(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        data = self.dev.rpc.get_config(model=True)
        self.assertEqual(data.tag, "data")

    def test_get_config_format_json_JSONLoadError_with_line(self):
        self.dev._conn.rpc = MagicMock(side_effect=self._mock_manager)
        self.dev.facts._cache["version_info"] = version_info("15.1X46-D15.3")
        try:
            self.dev.rpc.get_config(options={"format": "json"})
        except JSONLoadError as ex:
            self.assertTrue(
                re.search(
                    r"Expecting '?,'? delimiter: line 17 column 39 \(char 516\)",
                    ex.ex_msg,
                )
                is not None
            )

    def _mock_manager(self, *args, **kwargs):
        if kwargs:
            if "normalize" in kwargs and args:
                return self._read_file(args[0].tag + ".xml")
            device_params = kwargs["device_params"]
            device_handler = make_device_handler(device_params)
            session = SSHSession(device_handler)
            return Manager(session, device_handler)

        if args:
            if len(args[0]) > 0 and args[0][0].tag == "bgp":
                return self._read_file(args[0].tag + "_bgp_openconfig.xml")
            elif (
                args[0].attrib.get("format") == "json"
                and args[0].tag == "get-configuration"
            ):
                return self._read_file(args[0].tag + ".json")
            return self._read_file(args[0].tag + ".xml")

    def _read_file(self, fname):
        from ncclient.xml_ import NCElement

        fpath = os.path.join(os.path.dirname(__file__), "rpc-reply", fname)
        with open(fpath) as fp:
            foo = fp.read()
        return NCElement(foo, self.dev._conn._device_handler.transform_reply())



try:
    import unittest2 as unittest
except ImportError:
    import unittest
from jnpr.junos.utils.config import Config
import nose2
from unittest.mock import patch, MagicMock, call
import re
import sys
import os
from lxml import etree
import six
import socket

from jnpr.junos.console import Console
from jnpr.junos.transport.tty_netconf import tty_netconf
from jnpr.junos.transport.tty_telnet import Terminal


if sys.version < "3":
    builtin_string = "__builtin__"
else:
    builtin_string = "builtins"


class TestConsole(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Save object state
        cls.open = tty_netconf.open

    @classmethod
    def tearDownClass(cls):
        # Revert object state
        tty_netconf.open = cls.open

    @patch("jnpr.junos.transport.tty_telnet.Telnet._tty_open")
    @patch("jnpr.junos.transport.tty_telnet.telnetlib.Telnet.expect")
    @patch("jnpr.junos.transport.tty_telnet.Telnet.write")
    def setUp(self, mock_write, mock_expect, mock_open):
        tty_netconf.open = MagicMock()
        mock_expect.side_effect = [
            (1, re.search(r"(?P<login>ogin:\s*$)", "login: "), six.b("\r\r\n ogin:")),
            (
                2,
                re.search(r"(?P<passwd>assword:\s*$)", "password: "),
                six.b("\r\r\n password:"),
            ),
            (
                3,
                re.search(r"(?P<shell>%|#\s*$)", "junos % "),
                six.b("\r\r\nroot@device:~ # "),
            ),
        ]
        self.dev = Console(host="1.1.1.1", user="lab", password="lab123", mode="Telnet")
        self.dev.open()

    @patch("jnpr.junos.console.Console._tty_logout")
    def tearDown(self, mock_tty_logout):
        self.dev.close()

    def test_telnet_host_none(self):
        self.dev = Console(host=None, user="lab", password="lab123", mode="Telnet")
        self.assertTrue(self.dev.open()["failed"])

    @patch("jnpr.junos.console.warnings")
    def test_telnet_old_fact_warning(self, mock_warn):
        self.dev = Console(
            host="1.1.1.1",
            user="lab",
            password="lab123",
            mode="Telnet",
            fact_style="old",
        )
        mock_warn.assert_has_calls(
            [
                call.warn(
                    "fact-style old will be removed in a future release.",
                    RuntimeWarning,
                )
            ]
        )

    @patch("jnpr.junos.transport.tty_telnet.Telnet._tty_open")
    @patch("jnpr.junos.transport.tty_telnet.telnetlib.Telnet.expect")
    @patch("jnpr.junos.transport.tty_telnet.Telnet.write")
    def test_login_bad_password(self, mock_write, mock_expect, mock_open):
        tty_netconf.open = MagicMock()
        mock_expect.side_effect = [
            (1, re.search(r"(?P<login>ogin:\s*$)", "login: "), six.b("\r\r\n ogin:")),
            (
                2,
                re.search(r"(?P<passwd>assword:\s*$)", "password: "),
                six.b("\r\r\n password:"),
            ),
            (
                3,
                re.search("(?P<badpasswd>ogin incorrect)", "login incorrect"),
                six.b("\r\r\nlogin incorrect"),
            ),
        ]
        self.dev = Console(host="1.1.1.1", user="lab", password="lab123", mode="Telnet")
        self.assertRaises(StopIteration, self.dev.open)

    @patch("jnpr.junos.console.Console._tty_logout")
    @patch("jnpr.junos.transport.tty_telnet.Telnet._tty_open")
    @patch("jnpr.junos.transport.tty_telnet.telnetlib.Telnet.expect")
    @patch("jnpr.junos.transport.tty_telnet.Telnet.write")
    def test_with_context(self, mock_write, mock_expect, mock_open, mock_logout):
        tty_netconf.open = MagicMock()

        mock_expect.side_effect = [
            (1, re.search(r"(?P<login>ogin:\s*$)", "login: "), six.b("\r\r\n ogin:")),
            (
                2,
                re.search(r"(?P<passwd>assword:\s*$)", "password: "),
                six.b("\r\r\n password:"),
            ),
            (
                3,
                re.search(r"(?P<shell>%|#\s*$)", "junos % "),
                six.b("\r\r\nroot@device:~ # "),
            ),
        ]
        with Console(
            host="1.1.1.1", user="lab", password="lab123", mode="Telnet"
        ) as dev:
            self.assertTrue(isinstance(self.dev, Console))

    @patch("jnpr.junos.console.Console._tty_login")
    def test_console_open_error(self, mock_tty_login):
        mock_tty_login.side_effect = RuntimeError
        self.assertRaises(RuntimeError, self.dev.open)

    def test_console_connected(self):
        self.assertTrue(self.dev.connected)
        self.assertFalse(self.dev._gather_facts)

    @patch("jnpr.junos.console.Console._tty_logout")
    def test_console_close_error(self, mock_logout):
        mock_logout.side_effect = RuntimeError
        self.assertRaises(RuntimeError, self.dev.close)

    @patch("jnpr.junos.console.Console._tty_logout")
    def test_console_close_socket_error(self, mock_logout):
        mock_logout.side_effect = socket.error
        self.assertRaises(socket.error, self.dev.close)

    @patch("jnpr.junos.console.Console._tty_logout")
    def test_console_close_socket_conn_reset(self, mock_logout):
        mock_logout.side_effect = socket.error("Connection reset by peer")
        self.dev.close()
        self.assertFalse(self.dev.connected)

    @patch("jnpr.junos.console.Console._tty_logout")
    def test_console_close_telnet_conn_closed(self, mock_logout):
        mock_logout.side_effect = EOFError("telnet connection closed")
        self.dev.close()
        self.assertFalse(self.dev.connected)

    @patch("jnpr.junos.transport.tty_telnet.Telnet")
    @patch("jnpr.junos.console.Console._tty_login")
    def test_console_tty_open_err(self, mock_login, mock_telnet):
        with patch(
            "jnpr.junos.transport.tty_telnet." "telnetlib.Telnet.open"
        ) as mock_open:
            mock_telnet.RETRY_OPEN = 1
            mock_login.side_effect = ValueError
            self.dev._tty.LOGIN_RETRY = self.dev._tty.RETRY_OPEN = 1
            self.assertRaises(ValueError, self.dev.open)

    @patch("jnpr.junos.transport.tty_serial.Serial._tty_open")
    @patch("jnpr.junos.transport.tty_serial.serial.Serial.readline")
    @patch("jnpr.junos.transport.tty_serial.Serial.write")
    def test_console_serial(self, mock_write, mock_expect, mock_open):
        tty_netconf.open = MagicMock()
        mock_expect.side_effect = [
            six.b("\r\r\n Login:"),
            six.b("\r\r\n password:"),
            six.b("\r\r\nroot@device:~ # "),
        ]
        self.dev = Console(host="1.1.1.1", user="lab", password="lab123", mode="serial")
        self.dev.open()
        self.assertTrue(self.dev.connected)
        self.assertFalse(self.dev._gather_facts)

    def test_wrong_mode(self):
        dev = Console(host="1.1.1.1", user="lab", password="lab123", mode="testing")
        self.assertRaises(AttributeError, dev.open)

    @patch("jnpr.junos.transport.tty_telnet.Telnet._tty_close")
    def test_console_close_error_skip_logout(self, mock_close):
        mock_close.side_effect = RuntimeError
        self.assertRaises(RuntimeError, self.dev.close, skip_logout=True)

    @patch("jnpr.junos.transport.tty_netconf.tty_netconf.rpc")
    def test_console_zeroize(self, mock_zeroize):
        self.dev.zeroize()
        self.assertTrue(mock_zeroize.called)

    @patch("jnpr.junos.transport.tty_netconf.tty_netconf.rpc")
    @patch("jnpr.junos.console.FACT_LIST")
    @patch("jnpr.junos.device.warnings")
    def test_console_gather_facts(self, mock_warnings, mock_fact_list, mock_rpc):
        self.dev._fact_style = "old"
        from jnpr.junos.ofacts.session import facts_session

        mock_fact_list.__iter__.return_value = [facts_session]
        self.dev.facts_refresh()
        self.assertEqual(mock_rpc.call_count, 8)

    @patch("jnpr.junos.console.Console._tty_login")
    @patch("jnpr.junos.console.FACT_LIST")
    @patch("jnpr.junos.device.warnings")
    def test_console_gather_facts_true(self, mock_warnings, mock_fact_list, tty_login):
        self.dev._fact_style = "old"
        self.dev.facts = self.dev.ofacts
        from jnpr.junos.ofacts.session import facts_session

        mock_fact_list.__iter__.return_value = [facts_session]
        self.dev._gather_facts = True
        self.dev.open()
        self.assertEqual(
            self.dev.facts,
            {
                "2RE": False,
                "RE_hw_mi": False,
                "ifd_style": "CLASSIC",
                "serialnumber": "UNKNOWN",
                "model": "UNKNOWN",
                "vc_capable": False,
                "switch_style": "NONE",
                "personality": "UNKNOWN",
            },
        )

    @patch("jnpr.junos.transport.tty.sleep")
    @patch("ncclient.operations.rpc.RPCReply.parse")
    @patch("jnpr.junos.transport.tty_telnet.telnetlib.Telnet.write")
    @patch("jnpr.junos.transport.tty_netconf.select.select")
    @patch("jnpr.junos.transport.tty_telnet.telnetlib.Telnet.read_until")
    def test_load_console(
        self, mock_read_until, mock_select, mock_write, mock_parse, mock_sleep
    ):
        mock_select.return_value = ([self.dev._tty._rx], [], [])
        xml = """<policy-options>
                  <policy-statement>
                    <name>F5-in</name>
                    <term>
                        <name>testing</name>
                        <then>
                            <accept/>
                        </then>
                    </term>
                    <from>
                        <protocol>mpls</protocol>
                    </from>
                </policy-statement>
                </policy-options>"""

        mock_read_until.return_value = six.b(
            """
        <rpc-reply xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" xmlns:junos="http://xml.juniper.net/junos/15.2I0/junos">
            <load-configuration-results>
            <ok/>
            </load-configuration-results>
            </rpc-reply>
            ]]>]]>"""
        )
        cu = Config(self.dev)
        op = cu.load(xml, format="xml")
        cu.commit()

    @patch("ncclient.operations.rpc.RPCReply.parse")
    @patch("jnpr.junos.transport.tty_netconf.tty_netconf._receive")
    @patch("jnpr.junos.transport.tty_telnet.Telnet.rawwrite")
    def test_console_rpc_call(self, mock_write, mock_rcv, mock_parse):
        self.dev._tty.nc.rpc = MagicMock(side_effect=self._mock_manager)
        op = self.dev.rpc.get_chassis_inventory()
        self.assertEqual(op.tag, "chassis-inventory")

    @patch("ncclient.operations.rpc.RPCReply.parse")
    @patch("jnpr.junos.transport.tty_netconf.tty_netconf._receive")
    @patch("jnpr.junos.transport.tty_telnet.Telnet.rawwrite")
    def test_console_rpc_call_exception(self, mock_write, mock_rcv, mock_parse):
        mock_rcv.return_value = "<output>testing</output>"
        op = self.dev.rpc.get_chassis_inventory()
        self.assertEqual(op.tag, "output")

    def test_timeout_getter_setter(self):
        self.dev.timeout = 1
        self.assertEqual(1, self.dev.timeout)

    # below 2 function will be used in future.
    def _mock_manager(self, *args, **kwargs):
        if args:
            return self._read_file(etree.XML(args[0]).tag + ".xml")

    def _read_file(self, fname):
        from ncclient.xml_ import NCElement

        fpath = os.path.join(os.path.dirname(__file__), "rpc-reply", fname)
        with open(fpath) as fp:
            return fp.read()



import unittest
import nose2
from jnpr.junos.exception import (
    RpcError,
    CommitError,
    ConnectError,
    ConfigLoadError,
    RpcTimeoutError,
    SwRollbackError,
    JSONLoadError,
)
from jnpr.junos import Device
from lxml import etree

__author__ = "Nitin Kumar, Rick Sherman"
__credits__ = "Jeremy Schulman"


commit_xml = """
        <rpc-error xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" xmlns:junos="http://xml.juniper.net/junos/12.1X46/junos" xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <error-severity>error</error-severity>
        <source-daemon>dcd</source-daemon>
        <error-path>[edit interfaces ge-0/0/1]</error-path>
        <error-info>
        <bad-element>unit 2</bad-element>
        </error-info>
        <error-message>
        Only unit 0 is valid for this encapsulation
        </error-message>
        </rpc-error>
    """

rpc_xml = """
    <rpc-error xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" xmlns:junos="http://xml.juniper.net/junos/12.1X47/junos" xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0">
    <error-severity>error</error-severity>
    <error-info>
    <bad-element>bgp</bad-element>
    </error-info>
    <error-message>syntax error</error-message>
    </rpc-error>
    """

conf_xml = """
    <rpc-error>
        <error-severity>error</error-severity>
        <error-info>
            <bad-element>system1</bad-element>
        </error-info>
        <error-message>syntax error</error-message>
    </rpc-error>
"""

multi_warning_xml = """
<rpc-reply xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" xmlns:junos="http://xml.juniper.net/junos/16.1I0/junos" xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0">
<load-configuration-results>
<rpc-error>
<error-severity>warning</error-severity>
<error-message>
statement not found
</error-message>
</rpc-error>
<rpc-error>
<error-severity>warning</error-severity>
<error-message>
statement not found
</error-message>
</rpc-error>
<ok/>
</load-configuration-results>
</rpc-reply>
"""

config_json = """{
    "configuration" : {
      "system" : {
            "services" : {
                "telnet" : [nul] 
             }
        }
    }
}"""


class Test_RpcError(unittest.TestCase):
    def test_rpcerror_repr(self):
        rsp = etree.XML(rpc_xml)
        obj = RpcError(rsp=rsp)
        err = "RpcError(severity: error, bad_element: bgp, " "message: syntax error)"
        self.assertEqual(str, type(obj.__repr__()))
        self.assertEqual(obj.__repr__(), err)

    def test_rpcerror_jxml_check(self):
        # this test is intended to hit jxml code
        rsp = etree.XML(commit_xml)
        obj = CommitError(rsp=rsp)
        self.assertEqual(obj.rpc_error["bad_element"], "unit 2")

    def test_ConnectError(self):
        self.dev = Device(host="1.1.1.1", user="rick")
        obj = ConnectError(self.dev)
        self.assertEqual(obj.user, "rick")
        self.assertEqual(obj.host, "1.1.1.1")
        self.assertEqual(obj.port, 830)
        self.assertEqual(repr(obj), "ConnectError(1.1.1.1)")

    def test_ConnectError_msg(self):
        self.dev = Device(host="1.1.1.1", user="rick")
        obj = ConnectError(self.dev, msg="underlying exception info")
        self.assertEqual(obj.msg, "underlying exception info")
        self.assertEqual(
            repr(obj), "ConnectError(host: 1.1.1.1, msg: underlying exception info)"
        )

    def test_CommitError_repr(self):
        rsp = etree.XML(commit_xml)
        obj = CommitError(rsp=rsp)
        err = (
            "CommitError(edit_path: [edit interfaces ge-0/0/1], "
            "bad_element: unit 2, message: Only unit 0 is valid "
            "for this encapsulation)"
        )
        self.assertEqual(obj.__repr__(), err)

    def test_ConfigLoadError_repr(self):
        rsp = etree.XML(conf_xml)
        obj = ConfigLoadError(rsp=rsp)
        err = (
            "ConfigLoadError(severity: error, bad_element: "
            "system1, message: syntax error)"
        )
        self.assertEqual(obj.__repr__(), err)

    def test_RpcTimeoutError_repr(self):
        dev = Device("test")
        obj = RpcTimeoutError(dev=dev, cmd="test", timeout=50)
        err = "RpcTimeoutError(host: test, cmd: test, timeout: 50)"
        self.assertEqual(obj.__repr__(), err)

    def test_SwRollbackError_repr(self):
        obj = SwRollbackError(rsp="Single RE exception")
        err = "SwRollbackError(output: Single RE exception)"
        self.assertEqual(obj.__repr__(), err)

    def test_SwRollbackError_repr_multi(self):
        obj = SwRollbackError(re="test1", rsp="Multi RE exception")
        err = "SwRollbackError(re: test1, output: Multi RE exception)"
        self.assertEqual(obj.__repr__(), err)

    def test_repr_multi_warning(self):
        rsp = etree.XML(multi_warning_xml)
        from ncclient.operations import RPCError

        warn_msg = """
        <rpc-error xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" xmlns:junos="http://xml.juniper.net/junos/16.1I0/junos" xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <error-severity>warning</error-severity>
        <error-message>
        statement not found
        </error-message>
        </rpc-error>"""
        errs = RPCError(etree.XML(warn_msg))
        errs.errors = [errs, errs]
        obj = RpcError(rsp=rsp, errs=errs)
        self.assertEqual(obj.rpc_error["severity"], "warning")

    def test_json_error(self):
        err = "ValueError: No JSON object could be decoded"
        obj = JSONLoadError(err, config_json)
        errs = "JSONLoadError(reason: ValueError: No JSON object could be decoded)"
        self.assertEqual(obj.__repr__(), errs)

    def test_json_error_offending_line(self):
        err = "ValueError: No"
        obj = JSONLoadError(err, config_json)
        obj.offending_line = "Value"
        errs = (
            "JSONLoadError(reason: ValueError: No, "
            "\nThe offending config appears to be: "
            "\nValue)"
        )
        self.assertEqual(obj.__repr__(), errs)



__author__ = "Nitin Kumar, Rick Sherman"
__credits__ = "Jeremy Schulman"

import unittest
from unittest.mock import patch
import nose2

from jnpr.junos import Device
from jnpr.junos.ofacts.ifd_style import facts_ifd_style as ifd_style


class TestIFDStyle(unittest.TestCase):
    @patch("jnpr.junos.device.warnings")
    def setUp(self, mock_warnings):
        self.facts = {}
        self.dev = Device(
            host="1.1.1.1",
            user="rick",
            password="password123",
            gather_facts=False,
            fact_style="old",
        )

    def test_ifd_style_if_condition(self):
        self.facts["personality"] = "SWITCH"
        ifd_style(self.dev, self.facts)
        self.assertEqual(self.facts["ifd_style"], "SWITCH")

    def test_ifd_style_else_condition(self):
        self.facts["personality"] = "TEXT"
        ifd_style(self.dev, self.facts)
        self.assertEqual(self.facts["ifd_style"], "CLASSIC")



__author__ = "jnpr-community-netdev"

try:
    import unittest2 as unittest
except ImportError:
    import unittest

import yaml


class test(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from jnpr.junos import Device

        with open("config.yaml", "r") as fyaml:
            cfg = yaml.safe_load(fyaml)
        self.dev = Device(**cfg)
        self.dev.open()

    @classmethod
    def tearDownClass(self):
        self.dev.close()

    def test_shell_run(self):
        from jnpr.junos.utils.start_shell import StartShell

        with StartShell(self.dev) as sh:
            output = sh.run("pwd")
            self.assertTrue(output[0])

    def test_shell_run_with_sleep(self):
        from jnpr.junos.utils.start_shell import StartShell

        with StartShell(self.dev) as sh:
            output = sh.run("hostname", sleep=2)
            self.assertTrue(output[0])

    def test_shell_run_shell_type_ssh(self):
        from jnpr.junos.utils.start_shell import StartShell

        with StartShell(self.dev, shell_type="ssh") as sh:
            output = sh.run("hostname", sleep=2)
            self.assertTrue(output[0])

    def test_shell_run_shell_type_csh(self):
        from jnpr.junos.utils.start_shell import StartShell

        with StartShell(self.dev, shell_type="csh") as sh:
            output = sh.run("hostname", sleep=2)
            self.assertTrue(output[0])



import unittest
import nose2

from jnpr.junos import Device
from jnpr.junos.utils.start_shell import StartShell

from unittest.mock import patch, MagicMock, call

__author__ = "Rick Sherman"
__credits__ = "Jeremy Schulman, Nitin Kumar"


class TestStartShell(unittest.TestCase):
    @patch("paramiko.SSHClient")
    def setUp(self, mock_connect):
        self.dev = Device(host="1.1.1.1")
        self.shell = StartShell(self.dev)

    @patch("paramiko.SSHClient")
    @patch("jnpr.junos.utils.start_shell.StartShell.wait_for")
    def test_startshell_open_with_shell_term(self, mock_wait, mock_connect):
        mock_wait.return_value = ["user # "]
        self.shell.open()
        mock_wait.assert_called_with("(%|>|#|\\$)")

    @patch("paramiko.SSHClient")
    @patch("jnpr.junos.utils.start_shell.StartShell.wait_for")
    def test_startshell_open_with_junos_term(self, mock_wait, mock_connect):
        mock_wait.return_value = ["user > "]
        self.shell.open()
        mock_wait.assert_called_with("(%|#|\\$)\\s")

    @patch("paramiko.SSHClient")
    @patch("jnpr.junos.utils.start_shell.StartShell.wait_for")
    def test_startshell_open_with_bourne_shell(self, mock_wait, mock_connect):
        mock_wait.return_value = ["foo@bar:~$ "]
        self.shell.open()
        mock_wait.assert_called_with("(%|>|#|\\$)")

    @patch("paramiko.SSHClient")
    def test_startshell_close(self, mock_connect):
        self.shell._chan = MagicMock()
        self.shell._client = MagicMock()
        self.shell.close()
        self.shell._client.close.assert_called_once()

    @patch("jnpr.junos.utils.start_shell.StartShell.wait_for")
    def test_startshell_run(self, mock_wait):
        self.shell._chan = MagicMock()
        mock_wait.return_value = ["user % "]
        self.shell.run("ls")
        self.assertTrue(call.send("echo $?") in self.shell._chan.mock_calls)

    @patch("jnpr.junos.utils.start_shell.select")
    def test_startshell_wait_for(self, mock_select):
        mock_select.return_value = ["> ", 2, 3]
        self.shell._chan = MagicMock()
        self.shell._chan.recv.return_value = "> "
        self.assertTrue(self.shell.wait_for("> ")[0].endswith("> "))

    @patch("jnpr.junos.utils.start_shell.select")
    def test_startshell_wait_for_regex(self, mock_select):
        mock_select.return_value = ["> ", 2, 3]
        self.shell._chan = MagicMock()
        # output from command: cli -c "show version"
        self.shell._chan.recv.return_value = """
        ------------
        JUNOS Services Deep Packet Inspection package [15.1
        ---(more)---
        """
        self.assertTrue(
            str(self.shell.wait_for(r"---\(more\s?\d*%?\)---\n\s*|%")[0])
            in self.shell._chan.recv.return_value
        )

    @patch("jnpr.junos.utils.start_shell.StartShell.open")
    @patch("jnpr.junos.utils.start_shell.StartShell.close")
    def test_startshell_context(self, mock_close, mock_open):
        with StartShell(self.dev) as shell:
            shell._chan = MagicMock()
            shell.send("test")
        mock_close.assert_called_once_with()

    @patch("jnpr.junos.utils.start_shell.StartShell.wait_for")
    def test_startshell_run_regex(self, mock_wait_for):
        self.shell._chan = MagicMock()
        mock_wait_for.return_value = [
            """
        ------------
        JUNOS Services Deep Packet Inspection package [15.1
        ---(more)---
        """
        ]
        self.assertTrue(
            self.shell.run("show version", r"---\(more\s?\d*%?\)---\n\s*|%")[0]
        )

    @patch("jnpr.junos.utils.start_shell.StartShell.wait_for")
    def test_startshell_run_this_None(self, mock_wait_for):
        self.shell._chan = MagicMock()
        mock_wait_for.return_value = [
            """
        ------------
        JUNOS Services Deep Packet Inspection package [15.1
        """
        ]
        self.assertTrue(self.shell.run("show version", this=None)[0])



import socket
import sys

try:
    import unittest2 as unittest
except ImportError:
    import unittest
import nose2
from unittest.mock import MagicMock, patch
from jnpr.junos.transport.tty_ssh import SSH


class TestTTYSSH(unittest.TestCase):
    @patch("jnpr.junos.transport.tty_ssh.paramiko")
    def setUp(self, mock_paramiko):
        self.ssh_conn = SSH(
            host="1.1.1.1", user="test", password="password123", port=3007, timeout=30
        )
        self._ssh = mock_paramiko.invoke_shell

    def test_open(self):
        self.ssh_conn._tty_open()
        self.ssh_conn._ssh_pre.connect.assert_called()

    def test_open_exception(self):
        self.ssh_conn._ssh_pre.connect.side_effect = socket.error
        SSH.RETRY_OPEN = 1
        SSH.RETRY_BACKOFF = 0.1
        self.assertRaises(RuntimeError, self.ssh_conn._tty_open)
        # reset
        SSH.RETRY_OPEN = 3
        SSH.RETRY_BACKOFF = 2

    def test_read(self):
        self.ssh_conn._tty_open()
        self.assertRaises(ValueError, self.ssh_conn.read)

    def test_close(self):
        self.assertEqual(self.ssh_conn._tty_close(), None)

    @patch("jnpr.junos.transport.tty_ssh.paramiko")
    def test_tty_ssh_baud(self, mock_paramiko):
        self.ssh_conn = SSH(
            host="1.1.1.1",
            user="test",
            password="password123",
            port=3007,
            timeout=30,
            baud=0,
        )

        self.ssh_conn._tty_open()
        self.ssh_conn.rawwrite("<rpc>")
        self.ssh_conn._ssh.sendall.assert_called_with("<rpc>")

    def test_tty_ssh_rawwrite_sys_py3(self):
        with patch.object(sys.modules["sys"], "version", "3.x") as mock_sys:
            self.ssh_conn._tty_open()
            content = MagicMock()
            self.ssh_conn.rawwrite(content)
            content.decode.assert_called_with("utf-8")



__author__ = "Rick Sherman, Nitin Kumar"
__credits__ = "Jeremy Schulman"

import unittest
import os
import sys

import nose2
import yaml

from jnpr.junos import Device

from ncclient.manager import Manager, make_device_handler
from ncclient.transport import SSHSession
from lxml import etree
from unittest.mock import MagicMock, patch

from jnpr.junos.factory import loadyaml
from jnpr.junos.factory.factory_loader import FactoryLoader

try:
    _YAML_ = loadyaml("lib/jnpr/junos/cfgro/srx")
except:
    # Try to load the template relative to test base
    try:
        _YAML_ = loadyaml(
            os.path.join(
                os.path.dirname(__file__), "../../..", "lib/jnpr/junos/cfgro/srx.yml"
            )
        )
    except:
        raise

globals().update(_YAML_)

yaml_data = """---
    UserTable:
      get: system/login/user
      required_keys:
        user: name
      view: userView

    userView:
      groups:
        auth: authentication
      fields:
        uid: uid
        class: class
        uidgroup: { uid: group }
        fullgroup: { full-name: group }
      fields_auth:
        pass: encrypted-password

    GroupTable:
        get: groups
        item:
        args_key: name
        options: {}
      """
globals().update(FactoryLoader().load(yaml.load(yaml_data, Loader=yaml.FullLoader)))

yaml_bgp_data = """---
  BgpTable:
    set: protocols/bgp/group
    key-field:
      - bgp_name
    view: BgpView

  BgpView:
    groups:
      neigh : neighbor
    fields:
      bgp_name   : { 'name' :
                        { 'type' : 'str', 'minValue' : 3, 'maxValue': 12 } }
      bgp_type   : {'type' :
                        {'type': { 'enum': ['external', 'internal'] },
                                    'default': 'external' } }
      local_addr : local-address
      peer       : { 'peer-as' :
                        { 'type' : 'int', 'minValue': 0, 'maxValue': 200} }
    fields_neigh:
      neigh      : name
   """

globals().update(FactoryLoader().load(yaml.load(yaml_bgp_data, Loader=yaml.FullLoader)))


@unittest.skipIf(sys.platform == "win32", "will work for windows in coming days")
class TestFactoryCfgTable(unittest.TestCase):
    @patch("ncclient.manager.connect")
    def setUp(self, mock_connect):
        mock_connect.side_effect = self._mock_manager
        self.dev = Device(
            host="1.1.1.1", user="rick", password="password123", gather_facts=False
        )
        self.dev.open()
        self.zit = ZoneIfsTable(self.dev)
        self.ut = UserTable(self.dev)
        self.bgp = BgpTable(self.dev)

    def test_cfgtable_path(self):
        fname = "user.xml"
        path = os.path.join(os.path.dirname(__file__), "rpc-reply", fname)
        ut = UserTable(path=path)
        ut.get()
        self.assertEqual(ut[0].uid, "2000")

    def test_cfgtable_xml(self):
        fname = "user.xml"
        xml = self._read_file(fname)
        ut = UserTable(xml=xml)
        ut.get()
        self.assertEqual(ut[0].uid, "2000")

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_junos(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.dev.ON_JUNOS = True
        self.ut.get(user="test")
        self.assertEqual(self.ut[0]["uidgroup"], "global")

    @patch("jnpr.junos.Device.execute")
    @patch("jnpr.junos.jxml.conf_transform")
    def test_cfgtable_junos1(self, mock_jxml, mock_execute):
        mock_execute.side_effect = self._mock_manager
        mock_jxml.result_value = None
        with patch.dict("sys.modules", junos=MagicMock()):
            import junos

            self.dev.ON_JUNOS = True
            self.bgp.get()
            self.assertEqual(self.bgp.bgp_type, "external")

    @patch("jnpr.junos.Device.execute")
    @patch("jnpr.junos.jxml.conf_transform")
    def test_cfgtable_junos2(self, mock_jxml, mock_execute):
        mock_execute.side_effect = self._mock_manager
        mock_jxml.result_value = None
        with patch.dict("sys.modules", junos=MagicMock()):
            import junos

            junos.Junos_Configuration = None
            self.dev.ON_JUNOS = True
            self.bgp.get()
            self.assertEqual(self.bgp.bgp_type, "external")

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_get(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.zit.get(security_zone="untrust")
        self.assertEqual(len(self.zit), 1)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_get_group(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.ut.get(user="test")
        self.assertEqual(self.ut[0]["uidgroup"], "global")

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_get_namesonly(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.zit.get(security_zone="untrust", namesonly=True)
        self.assertEqual(self.zit._get_cmd.xpath("//@recurse")[0], "false")

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_get_options(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.zit.get(
            security_zone="untrust", options={"inherit": "defaults", "groups": "groups"}
        )
        self.assertEqual(self.zit._get_opt, {"inherit": "defaults", "groups": "groups"})

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_table_options(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        gt = GroupTable(self.dev)
        gt.get()
        self.assertEqual(gt._get_opt, {})

    def test_optable_get_key_required_error(self):
        self.assertRaises(ValueError, self.zit.get)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_get_caller_provided_kvargs(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.zit.get(security_zone="untrust", key="test")
        self.assertTrue(mock_execute.called)

    def test_cfgtable_key_value_none(self):
        self.assertRaises(ValueError, self.zit.get, securityzone="untrust")

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_get_fields(self, mock_execute):
        mock_execute.side_effect = self._mock_manager
        self.zit._data_dict["get_fields"] = ["system-services"]
        self.zit.get(security_zone="untrust", key="host-inbound-traffic")
        self.assertTrue("get_fields" in self.zit._data_dict)

    def test_cfgtable_dot_none_RuntimeError(self):
        ret_val = (
            "<configuration><security><zones><test-zone>"
            '<interfaces recurse="false"/></test-zone></zones>'
            "</security></configuration>"
        )
        self.zit._buildxml = MagicMock(return_value=etree.fromstring(ret_val))
        self.assertRaises(RuntimeError, self.zit.get, security_zone="abc")

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_set(self, mock_execute):
        self.bgp.rpc.lock_configuration = MagicMock()
        self.bgp.bgp_name = "external_1"
        self.bgp.append()
        self.bgp.set()
        xml = self.bgp.get_table_xml()
        self.assertEqual(xml.xpath("protocols/bgp/group/name")[0].text, "external_1")

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_set_inactive(self, mock_execute):
        yaml_auto_data = """---
           UserConfigTable1:
             set: system/login
             key-field:
               - username
             view: UserConfigView1

           UserConfigView1:
              groups:  
                auth: authentication
              fields:
                user: user
                username: user/name
                classname: { user/class : { 'type' : { 'enum' : ['operator', 'read-only', 'super-user'] } } }
                uid: { user/uid : { 'type' : 'int', 'minValue' : 100, 'maxValue' : 64000 } }
              fields_auth:
                password: user/encrypted-password
           """
        globals().update(
            FactoryLoader().load(yaml.load(yaml_auto_data, Loader=yaml.FullLoader))
        )
        at = UserConfigTable1(self.dev)
        at.rpc.lock_configuration = MagicMock()
        at.username = "user1"
        at.user = {"inactive": "inactive"}
        at.append()
        at.set()
        xml = at.get_table_xml()
        self.assertEqual(
            xml.xpath('system/login/user[@inactive="inactive"]/name')[0].text, "user1"
        )

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_set_bool(self, mock_execute):
        yaml_auto_data = """---
           UserConfigTable1:
             set: system/login
             key-field:
               - username
             view: UserConfigView1

           UserConfigView1:
              groups:
                auth: authentication
              fields:
                user: user
                username: user/name
                classname: { user/class : { 'type' : { 'enum' : ['operator', 'read-only', 'super-user'] } } }
                uid: { user/uid : { 'type' : bool , 'default' :False} }
              fields_auth:
                password: user/encrypted-password
           """
        globals().update(
            FactoryLoader().load(yaml.load(yaml_auto_data, Loader=yaml.FullLoader))
        )
        at = UserConfigTable1(self.dev)
        at.rpc.lock_configuration = MagicMock()
        at.username = True
        at.user = {"inactive": "inactive"}
        at.append()
        at.set()
        xml = at.get_table_xml()
        self.assertNotEqual(xml.xpath("system/login/user/name")[0].text, "user1")

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_load(self, mock_execute):
        self.bgp.rpc.lock_configuration = MagicMock()
        self.bgp.bgp_name = "external_1"
        self.bgp.append()
        self.bgp.bgp_name = "external_2"
        self.bgp.append()
        self.bgp.load()
        xml = self.bgp.get_table_xml()
        self.assertEqual(xml.xpath("protocols/bgp/group/name")[0].text, "external_1")
        self.assertEqual(xml.xpath("protocols/bgp/group/name")[1].text, "external_2")

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_set_group(self, mock_execute):
        self.bgp.rpc.lock_configuration = MagicMock()
        self.bgp.bgp_name = "external_1"
        self.bgp.neigh = ["30.30.10.10"]
        self.bgp.bgp_type = "external"
        self.bgp.append()
        self.bgp.set()
        xml = self.bgp.get_table_xml()
        self.assertEqual(xml.xpath("protocols/bgp/group/name")[0].text, "external_1")
        self.assertEqual(
            xml.xpath("protocols/bgp/group/neighbor/name")[0].text, "30.30.10.10"
        )
        self.assertEqual(xml.xpath("protocols/bgp/group/type")[0].text, "external")

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_dict_set(self, mock_execute):
        self.bgp.rpc.lock_configuration = MagicMock()
        self.bgp["bgp_name"] = "external_3"
        self.bgp.append()
        self.bgp.set()
        xml = self.bgp.get_table_xml()
        self.assertEqual(xml.xpath("protocols/bgp/group/name")[0].text, "external_3")

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_invalid_field_error(self, mock_execute):
        def invalid_field():
            self.bgp.xyz = "bad"

        self.assertRaises(ValueError, invalid_field)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_invalid_field_dict_error(self, mock_execute):
        def invalid_field_dict():
            self.bgp["xyz"] = "bad"

        self.assertRaises(ValueError, invalid_field_dict)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_set_no_key_error(self, mock_execute):
        self.assertRaises(ValueError, self.bgp.append)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_enum_error(self, mock_execute):
        self.bgp.bgp_name = "external_1"
        self.bgp.bgp_type = "abc"
        self.assertRaises(ValueError, self.bgp.append)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_str_min_max_value_error(self, mock_execute):
        self.bgp.bgp_name = "ex"
        self.assertRaises(ValueError, self.bgp.append)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_int_min_max_value_error(self, mock_execute):
        self.bgp.bgp_name = "external_1"
        self.bgp.peer = 300
        self.assertRaises(ValueError, self.bgp.append)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_invalid_field_value_error(self, mock_execute):
        self.bgp.bgp_name = "external_1"
        self.bgp.peer = [[100]]
        self.assertRaises(ValueError, self.bgp.append)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_str_key_field(self, mock_execute):
        yaml_auto_data = """---
          AutoSysTable:
            set: routing-options/autonomous-system
            key-field:
              as_num
            view: AutoSysView

          AutoSysView:
            fields:
              as_num: as-number
           """
        globals().update(
            FactoryLoader().load(yaml.load(yaml_auto_data, Loader=yaml.FullLoader))
        )
        at = AutoSysTable(self.dev)
        at.rpc.lock_configuration = MagicMock()
        at.as_num = 100
        at.append()
        at.set()
        xml = at.get_table_xml()
        self.assertEqual(
            xml.xpath("routing-options/autonomous-system/as-number")[0].text, "100"
        )

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_field_value_xpath(self, mock_execute):
        yaml_auto_data = """---
          AutoSysTable:
            set: routing-options
            key-field:
              as_num
            view: AutoSysView

          AutoSysView:
            fields:
              as_num: autonomous-system/as-number
           """
        globals().update(
            FactoryLoader().load(yaml.load(yaml_auto_data, Loader=yaml.FullLoader))
        )
        at = AutoSysTable(self.dev)
        at.rpc.lock_configuration = MagicMock()
        at.as_num = 150
        at.append()
        at.set()
        xml = at.get_table_xml()
        self.assertEqual(
            xml.xpath("routing-options/autonomous-system/as-number")[0].text, "150"
        )

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_user_defined_type_error(self, mock_execute):
        yaml_auto_data = """---
          AutoSysTable:
            set: routing-options/autonomous-system
            key-field:
              - as_num
            view: AutoSysView

          AutoSysView:
            fields:
              as_num: {'as-number': {'type': {'UserDefined': ''}}}
           """
        globals().update(
            FactoryLoader().load(yaml.load(yaml_auto_data, Loader=yaml.FullLoader))
        )
        at = AutoSysTable(self.dev)
        at.rpc.lock_configuration = MagicMock()
        at.as_num = 100
        self.assertRaises(TypeError, at.append)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_wrong_type_error(self, mock_execute):
        yaml_auto_data = """---
          AutoSysTable:
            set: routing-options/autonomous-system
            key-field:
              - as_num
            view: AutoSysView

          AutoSysView:
            fields:
              as_num: {'as-number' : { 'type' : 'int'} }
           """
        globals().update(
            FactoryLoader().load(yaml.load(yaml_auto_data, Loader=yaml.FullLoader))
        )
        at = AutoSysTable(self.dev)
        at.as_num = "100"
        self.assertRaises(TypeError, at.append)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_unsupported_type_error(self, mock_execute):
        yaml_auto_data = """---
          AutoSysTable:
            set: routing-options/autonomous-system
            key-field:
              - as_num
            view: AutoSysView

          AutoSysView:
            fields:
              as_num: {'as-number' : { 'type' : 'interger'} }
           """
        globals().update(
            FactoryLoader().load(yaml.load(yaml_auto_data, Loader=yaml.FullLoader))
        )
        at = AutoSysTable(self.dev)
        at.as_num = 100
        self.assertRaises(TypeError, at.append)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_enum_value_str_error(self, mock_execute):
        yaml_auto_data = """---
          AutoSysTable:
            set: routing-options/autonomous-system
            key-field:
              - as_num
            view: AutoSysView

          AutoSysView:
            fields:
              as_num: {'as-number' : {'type' : {'enum': '100'}}}
           """
        globals().update(
            FactoryLoader().load(yaml.load(yaml_auto_data, Loader=yaml.FullLoader))
        )
        at = AutoSysTable(self.dev)
        at.as_num = 100
        self.assertRaises(ValueError, at.append)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_enum_value_type_error(self, mock_execute):
        yaml_auto_data = """---
          AutoSysTable:
            set: routing-options/autonomous-system
            key-field:
              - as_num
            view: AutoSysView

          AutoSysView:
            fields:
              as_num: {'as-number' : {'type' : {'enum': {'100': ''}}}}
           """
        globals().update(
            FactoryLoader().load(yaml.load(yaml_auto_data, Loader=yaml.FullLoader))
        )
        at = AutoSysTable(self.dev)
        at.as_num = 100
        self.assertRaises(TypeError, at.append)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_invalid_type_error(self, mock_execute):
        yaml_auto_data = """---
          AutoSysTable:
            set: routing-options/autonomous-system
            key-field:
              - as_num
            view: AutoSysView

          AutoSysView:
            fields:
              as_num: {'as-number': {'type': ['abc']}}
           """
        globals().update(
            FactoryLoader().load(yaml.load(yaml_auto_data, Loader=yaml.FullLoader))
        )
        at = AutoSysTable(self.dev)
        at.as_num = 100
        self.assertRaises(TypeError, at.append)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_invalid_key_field_type_error(self, mock_execute):
        yaml_auto_data = """---
          AutoSysTable:
            set: routing-options/autonomous-system
            key-field:
              as_num : as-number
            view: AutoSysView

          AutoSysView:
            fields:
              as_num: as-number
           """
        globals().update(
            FactoryLoader().load(yaml.load(yaml_auto_data, Loader=yaml.FullLoader))
        )
        self.assertRaises(TypeError, AutoSysTable, self.dev)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_invalid_key_field_not_defined_error(self, mock_execute):
        yaml_auto_data = """---
          AutoSysTable:
            set: routing-options/autonomous-system
            view: AutoSysView

          AutoSysView:
            fields:
              as_num: as-number
           """
        globals().update(
            FactoryLoader().load(yaml.load(yaml_auto_data, Loader=yaml.FullLoader))
        )
        self.assertRaises(ValueError, AutoSysTable, self.dev)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_invalid_view_not_defined_type_error(self, mock_execute):
        yaml_auto_data = """---
          AutoSysTable:
            set: routing-options/autonomous-system
           """
        globals().update(
            FactoryLoader().load(yaml.load(yaml_auto_data, Loader=yaml.FullLoader))
        )
        self.assertRaises(ValueError, AutoSysTable, self.dev)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_set_append_not_call_error(self, mock_execute):
        self.bgp.rpc.lock_configuration = MagicMock()
        self.bgp["bgp_name"] = "external_3"
        self.assertRaises(RuntimeError, self.bgp.set)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_load_append_not_call_error(self, mock_execute):
        self.bgp["bgp_name"] = "external_3"
        self.assertRaises(RuntimeError, self.bgp.load)

    @patch("jnpr.junos.Device.execute")
    def test_cfgtable_unfreeze(self, mock_execute):
        self.bgp._unfreeze()
        self.assertEqual(self.bgp._CfgTable__isfrozen, False)

    @patch("jnpr.junos.Device.execute")
    @patch("jnpr.junos.utils.config.Config.lock")
    @patch("jnpr.junos.utils.config.Config.unlock")
    def test_cfgtable_with_block(self, mock_execute, mock_unlock, mock_lock):
        with BgpTable(self.dev, mode="exclusive") as bgp:
            bgp.rpc.lock_configuration = MagicMock()
            bgp.bgp_name = "external_1"
            bgp.append()
            bgp.load()
        self.assertTrue(mock_lock.called and mock_unlock.called)

    def _read_file(self, fname):
        from ncclient.xml_ import NCElement

        fpath = os.path.join(os.path.dirname(__file__), "rpc-reply", fname)

        foo = open(fpath).read()

        if fname == "user.xml":
            return etree.fromstring(foo)

        rpc_reply = NCElement(
            foo, self.dev._conn._device_handler.transform_reply()
        )._NCElement__doc[0]
        return rpc_reply

    def _mock_manager(self, *args, **kwargs):
        if kwargs:
            device_params = kwargs["device_params"]
            device_handler = make_device_handler(device_params)
            session = SSHSession(device_handler)
            return Manager(session, device_handler)

        if args:
            if args[0].xpath("//configuration/system/login/user"):
                return self._read_file("get-configuration-user.xml")
            else:
                return self._read_file(args[0].tag + ".xml")



import logging

try:
    import unittest2 as unittest
except ImportError:
    import unittest

import nose2
from unittest.mock import MagicMock, patch

from jnpr.junos.transport.tty import Terminal
from jnpr.junos import exception as EzErrors


class TestTTY(unittest.TestCase):
    def setUp(self):
        logging.getLogger("jnpr.junos.tty")
        self.terminal = Terminal(user="test", password="password123", attempts=1)

    def test_login_bad_password_runtimeerror(self):
        self.terminal._badpasswd = 4
        self.terminal.read_prompt = MagicMock()
        self.terminal.read_prompt.return_value = (None, "badpasswd")
        self.terminal.write = MagicMock()
        self.assertRaises(RuntimeError, self.terminal._login_state_machine)

    def test_login_bad_password_ConnectAuthError(self):
        self.terminal._badpasswd = 1
        self.terminal.read_prompt = MagicMock()
        self.terminal.read_prompt.return_value = (None, "badpasswd")
        self.terminal.write = MagicMock()
        self.terminal._tty_close = MagicMock()
        self.assertRaises(EzErrors.ConnectAuthError, self.terminal._login_state_machine)

    @patch("jnpr.junos.transport.tty.sleep")
    def test_tty_no_login(self, mock_sleep):
        self.terminal._badpasswd = 4
        self.terminal.read_prompt = MagicMock()
        self.terminal.read_prompt.return_value = (None, "testing")
        self.terminal.write = MagicMock()
        self.assertRaises(RuntimeError, self.terminal._login_state_machine)
        self.terminal.write.assert_called_with("\n")
        self.terminal.read_prompt.return_value = (None, "testing")
        self.terminal.write = MagicMock()
        self.assertRaises(RuntimeError, self.terminal._login_state_machine)
        self.terminal.write.assert_called_with("<close-session/>")

    @patch("jnpr.junos.transport.tty.sleep")
    @patch("jnpr.junos.transport.tty.logger")
    def test_tty_cli(self, mock_logger, mock_sleep):
        self.terminal.read_prompt = MagicMock()
        self.terminal.read_prompt.return_value = (None, "cli")
        self.terminal._login_state_machine()
        self.assertEqual(self.terminal.state, 4)

    @patch("jnpr.junos.transport.tty.sleep")
    def test_loader(self, mock_sleep):
        self.terminal.write = MagicMock()
        self.terminal.read_prompt = MagicMock()
        self.terminal._loader = 2
        # self.terminal._login_state_machine = MagicMock()
        self.terminal.read_prompt.side_effect = [(None, "loader"), (None, "testing")]
        self.assertRaises(RuntimeError, self.terminal._login_state_machine)
        mock_sleep.assert_called_with(300)

    @patch("jnpr.junos.transport.tty.logger")
    def test_ev_shell(self, mock_logger):
        self.terminal.write = MagicMock()
        self.terminal.read_prompt = MagicMock()
        self.terminal.read_prompt.return_value = (None, "shell")
        self.terminal._login_state_machine()
        self.assertEqual(self.terminal.state, 4)

    @patch("jnpr.junos.transport.tty.sleep")
    def test_ev_option(self, mock_sleep):
        self.terminal.write = MagicMock()
        self.terminal.read_prompt = MagicMock()
        self.terminal.read_prompt.return_value = (None, "option")
        self.assertRaises(RuntimeError, self.terminal._login_state_machine)
        self.terminal.write.assert_called_with("1")
        self.assertEqual(self.terminal.state, 7)

    @patch("jnpr.junos.transport.tty.sleep")
    def test_tty_ev_netconf_closed(self, mock_sleep):
        self.terminal.write = MagicMock()
        self.terminal._tty_close = MagicMock()
        self.terminal.read_prompt = MagicMock()
        self.terminal.read_prompt.side_effect = iter(
            [(None, "netconf_closed"), (None, "shell"), (None, "login")]
        )
        self.assertTrue(self.terminal._logout_state_machine())

    @patch("jnpr.junos.transport.tty.sleep")
    def test_tty_already_logout(self, mock_sleep):
        self.terminal.write = MagicMock()
        self.terminal.read_prompt = MagicMock()
        self.terminal.read_prompt.return_value = (None, None)
        self.assertTrue(self.terminal._logout_state_machine())

    @patch("jnpr.junos.transport.tty.sleep")
    def test_tty_login_state_machine_loader(self, mock_sleep):
        self.terminal.write = MagicMock()
        self.terminal._loader = 1
        self.terminal.read_prompt = MagicMock()
        self.terminal.read_prompt.side_effect = iter(
            [(None, "loader"), (None, "hotkey"), (None, "shell")]
        )
        self.assertRaises(RuntimeError, self.terminal._login_state_machine)

    @patch("jnpr.junos.transport.tty.sleep")
    def test_tty_login_state_machine_loader(self, mock_sleep):
        self.terminal.write = MagicMock()
        self.terminal._loader = 1
        self.terminal.read_prompt = MagicMock()
        self.terminal.read_prompt.side_effect = iter(
            [(None, "loader"), (None, "shell")]
        )
        try:
            self.terminal._login_state_machine()
        except RuntimeError as ex:
            self.assertEqual(str(ex), "probably corrupted image, stuck in loader")

    @patch("jnpr.junos.transport.tty.sleep")
    def test_tty_login_state_machine_hotkey(self, mock_sleep):
        self.terminal.write = MagicMock()
        self.terminal.read_prompt = MagicMock()
        self.terminal.read_prompt.return_value = (None, "hotkey")
        self.assertRaises(RuntimeError, self.terminal._login_state_machine)
        self.assertEqual(self.terminal.state, 8)  # 8 is for hot keys

    @patch("jnpr.junos.transport.tty.sleep")
    def test_tty_ev_tty_nologin(self, mock_sleep):
        self.terminal.write = MagicMock()
        self.terminal.read_prompt = MagicMock()
        self.terminal.read_prompt.return_value = (None, None)
        self.terminal.console_has_banner = True
        self.assertRaises(RuntimeError, self.terminal._login_state_machine)
        self.terminal.write.assert_called_with("\n")

    @patch("jnpr.junos.transport.tty.sleep")
    def test_tty_logout_state_machine_attempt_10(self, mock_sleep):
        self.terminal.write = MagicMock()
        self.terminal.read_prompt = MagicMock()
        self.terminal.read_prompt.return_value = (None, "cli")
        try:
            self.terminal._logout_state_machine()
        except RuntimeError as ex:
            self.assertEqual(str(ex), "logout_sm_failure")



try:
    import unittest2 as unittest
except ImportError:
    import unittest
import nose2
from unittest.mock import MagicMock, patch
import sys
import six

from jnpr.junos.console import Console

if sys.version < "3":
    builtin_string = "__builtin__"
else:
    builtin_string = "builtins"


class TestSerial(unittest.TestCase):
    @patch("jnpr.junos.transport.tty_serial.serial.Serial.open")
    @patch("jnpr.junos.transport.tty_serial.serial.Serial.write")
    @patch("jnpr.junos.transport.tty_serial.serial.Serial.flush")
    @patch("jnpr.junos.transport.tty_serial.Serial.read_prompt")
    @patch("jnpr.junos.transport.tty.tty_netconf.open")
    def setUp(self, mock_nc_open, mock_read, mock_flush, mock_write, mock_open):
        self.dev = Console(port="USB/ttyUSB0", baud=9600, mode="Serial")
        mock_read.side_effect = [
            ("login", "login"),
            ("passwd", "passwd"),
            ("shell", "shell"),
        ]
        self.dev.open()

    @patch("jnpr.junos.transport.tty.sleep")
    @patch("jnpr.junos.transport.tty.tty_netconf.close")
    @patch("jnpr.junos.transport.tty_serial.Serial.read_prompt")
    @patch("jnpr.junos.transport.tty_serial.Serial.write")
    @patch("jnpr.junos.transport.tty_serial.Serial._tty_close")
    def tearDown(
        self, mock_serial_close, mock_write, mock_read, mock_close, mock_sleep
    ):
        mock_read.side_effect = [
            ("shell", "shell"),
            ("login", "login"),
            ("cli", "cli"),
        ]
        self.dev.close()

    def test_console_connected(self):
        self.assertTrue(self.dev.connected)

    def test_close_connection(self):
        self.dev._tty._ser = MagicMock()
        self.dev.close(skip_logout=True)
        self.assertTrue(self.dev._tty._ser.close.called)

    @patch("jnpr.junos.transport.tty_serial.serial.Serial.open")
    def test_tty_serial_open_exception(self, mock_open):
        dev = Console(port="USB/ttyUSB0", baud=9600, mode="Serial")
        mock_open.side_effect = OSError
        self.assertRaises(RuntimeError, dev.open)

    def test_tty_serial_rawwrite(self):
        self.dev._tty._ser = MagicMock()
        self.dev._tty.rawwrite("test")
        self.dev._tty._ser.write.assert_called_with("test")

    def test_tty_serial_read(self):
        self.dev._tty._ser = MagicMock()
        self.dev._tty.read()
        self.dev._tty._ser.readline.assert_called()

    def test_tty_serial_read_prompt(self):
        self.dev._tty._ser = MagicMock()
        self.dev._tty.EXPECT_TIMEOUT = 0.1
        self.dev._tty._ser.readline.side_effect = ["", "test"]
        self.assertEqual(self.dev._tty.read_prompt()[0], None)


class TestSerialWin(unittest.TestCase):
    @patch("jnpr.junos.transport.tty_serial.serial.Serial.open")
    @patch("jnpr.junos.transport.tty_serial.serial.Serial.read")
    @patch("jnpr.junos.transport.tty_serial.serial.Serial.write")
    @patch("jnpr.junos.transport.tty_serial.serial.Serial.flush")
    @patch("jnpr.junos.transport.tty_serial.Serial.read_prompt")
    def setUp(self, mock_read, mock_flush, mock_write, mock_serial_read, mock_open):
        self.dev = Console(port="COM4", baud=9600, mode="Serial")
        mock_read.side_effect = [
            ("login", "login"),
            ("passwd", "passwd"),
            ("shell", "shell"),
        ]
        mock_serial_read.side_effect = [
            six.b(
                "<!-- No zombies were killed during the creation of this user interface -->"
            ),
            six.b(""),
            six.b(
                """<!-- user root, class super-user -->
<hello xmlns="urn:ietf:params:xml:ns:netconf:base:1.0">
  <capabilities>
    <capability>urn:ietf:params:netconf:base:1.0</capability>
    <capability>urn:ietf:params:netconf:capability:candidate:1.0</capability>
    <capability>urn:ietf:params:netconf:capability:confirmed-commit:1.0</capability>
    <capability>urn:ietf:params:netconf:capability:validate:1.0</capability>
    <capability>urn:ietf:params:netconf:capability:url:1.0?scheme=http,ftp,file</capability>
    <capability>urn:ietf:params:xml:ns:netconf:base:1.0</capability>
    <capability>urn:ietf:params:xml:ns:netconf:capability:candidate:1.0</capability>
    <capability>urn:ietf:params:xml:ns:netconf:capability:confirmed-commit:1.0</capability>
    <capability>urn:ietf:params:xml:ns:netconf:capability:validate:1.0</capability>
    <capability>urn:ietf:params:xml:ns:netconf:capability:url:1.0?scheme=http,ftp,file</capability>
    <capability>urn:ietf:params:xml:ns:yang:ietf-netconf-monitoring</capability>
    <capability>http://xml.juniper.net/netconf/junos/1.0</capability>
    <capability>http://xml.juniper.net/dmi/system/1.0</capability>
  </capabilities>
  <session-id>7478</session-id>
</hello>
]]>]]>"""
            ),
            six.b(""),
        ]
        self.dev.open()

    @patch("jnpr.junos.transport.tty.sleep")
    @patch("jnpr.junos.transport.tty.tty_netconf.close")
    @patch("jnpr.junos.transport.tty_serial.Serial.read_prompt")
    @patch("jnpr.junos.transport.tty_serial.Serial.write")
    @patch("jnpr.junos.transport.tty_serial.Serial._tty_close")
    def tearDown(
        self, mock_serial_close, mock_write, mock_read, mock_close, mock_sleep
    ):
        mock_read.side_effect = [
            ("shell", "shell"),
            ("login", "login"),
            ("cli", "cli"),
        ]
        self.dev.close()

    def test_tty_serial_win_connected(self):
        self.assertTrue(self.dev.connected)

    @patch("jnpr.junos.transport.tty.tty_netconf.close")
    @patch("jnpr.junos.transport.tty_serial.Serial._tty_close")
    def test_tty_serial_win_rpc_call(self, mock_serial_close, mock_close):
        self.dev._tty.read = MagicMock()
        self.dev._tty.rawwrite = MagicMock()
        self.dev._tty.read.side_effect = [
            six.b(
                '<rpc-reply xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"'
                ' xmlns:junos="http://xml.juniper.net/junos/15.1X49/junos">'
                '<route-engine-information xmlns="http://xml.juniper.net/ju'
                'nos/15.1X49/junos-chassis"><route-engine><status>OK</statu'
                's><temperature junos:celsius="45">45 degrees C / 113 degre'
                'es F</temperature><cpu-temperature junos:celsius="61">61 d'
                "egrees C / 141 degrees F</cpu-temperature><memory-system-t"
                "otal>4096</memory-system-total><memory-system-total-used>1"
                "024</memory-system-total-used><memory-system-total-util>25"
                "</memory-system-total-util><memory-control-plane>2624</mem"
                "ory-control-plane><memory-control-plane-used>682</memory-c"
                "ontrol-plane-used><memory-control-plane-util>26</memory-co"
                "ntrol-plane-util><memory-data-plane>1472</memory-data-plan"
                "e><memory-data-plane-used>353</memory-data-plane-used><mem"
                "ory-data-plane-util>24</memory-data-plane-util><cpu-user>1"
                "2</cpu-user><cpu-background>0</cpu-background><cpu-system>"
                "6</cpu-system><cpu-interrupt>0</cpu-interrupt><cpu-idle>83"
                "</cpu-idle><model>RE-SRX300</model><serial-number>CV0918AF"
                '1022</serial-number><start-time junos:seconds="1584539305"'
                ">2020-03-18 08:48:25 CDT</start-time><up-time junos:second"
                's="137925">1 day, 14 hours, 18 minutes, 45 seconds</up-tim'
                "e><last-reboot-reason>0x1:power cycle/failure</last-reboot"
                "-reason><load-average-one>0.12</load-average-one><load-ave"
                "rage-five>0.08</load-average-five><load-average-fifteen>0."
                "06</load-average-fifteen></route-engine></route-engine-inf"
                "ormation></rpc-reply>]]>]]>"
            )
        ]
        res = self.dev.rpc.get_route_engine_information()
        self.assertEqual(res.tag, "route-engine-information")



import os
import unittest
from io import StringIO
import nose2
from unittest.mock import patch
from jnpr.junos.jxml import (
    NAME,
    INSERT,
    remove_namespaces,
    cscript_conf,
    normalize_xslt,
)
from lxml import etree
from ncclient.xml_ import NCElement

__author__ = "Nitin Kumar, Rick Sherman"
__credits__ = "Jeremy Schulman"


class Test_JXML(unittest.TestCase):
    def test_name(self):
        op = NAME("test")
        self.assertEqual(op["name"], "test")

    def test_insert(self):
        op = INSERT("test")
        self.assertEqual(op["insert"], "test")

    def test_remove_namespaces(self):
        xmldata = """<xsl:stylesheet xmlns:xsl="http://xml.juniper.net/junos">
                    <xsl:template>
                        <!-- Handle comments properly -->
                        <xsl:attribute name="{myname}">
                        </xsl:attribute>
                    </xsl:template>
                </xsl:stylesheet>"""
        import xml.etree.ElementTree as ET

        parser = ET.XMLParser()
        root = ET.parse(StringIO(xmldata), parser)
        test = remove_namespaces(root)
        for elem in test.iter():
            i = elem.tag.find("}")
            if i > 0:
                i = i + 1
        self.assertTrue(i <= 0)

    def test_cscript_conf(self):
        op = cscript_conf(self._read_file("get-configuration.xml"))
        self.assertTrue(isinstance(op, etree._Element))

    @patch("ncclient.manager.make_device_handler")
    def test_cscript_conf_return_none(self, dev_handler):
        dev_handler.side_effects = ValueError
        op = cscript_conf(self._read_file("get-configuration.xml"))
        self.assertTrue(op is None)

    def test_cscript_conf_output_tag_child_element(self):
        xmldata = """<rpc-reply message-id="urn:uuid:932d11dc-9ae5-4d25-81fa-8b50ea2d3a03" xmlns:junos="http://xml.juniper.net/junos/19.3R0/junos">
  <output>
    shutdown: [pid 8683]
    Shutdown NOW!
  </output>
</rpc-reply>
"""
        xmldata_without_ns = """<rpc-reply message-id="urn:uuid:932d11dc-9ae5-4d25-81fa-8b50ea2d3a03">
  <output>
    shutdown: [pid 8683]
    Shutdown NOW!
  </output>
</rpc-reply>
"""
        rpc_reply = NCElement(xmldata, normalize_xslt.encode("UTF-8"))
        self.assertEqual(str(rpc_reply), xmldata_without_ns)

    def test_cscript_conf_output_tag_not_first_child_element(self):
        xmldata = """<rpc-reply message-id="urn:uuid:932d11dc-9ae5-4d25-81fa-8b50ea2d3a03" xmlns:junos="http://xml.juniper.net/junos/19.3R0/junos">
  <xyz>
    <output>
      shutdown: [pid 8683]
      Shutdown NOW!
    </output>
  </xyz>
</rpc-reply>
"""
        xmldata_without_ns = """<rpc-reply message-id="urn:uuid:932d11dc-9ae5-4d25-81fa-8b50ea2d3a03">
  <xyz>
    <output>shutdown: [pid 8683] Shutdown NOW!</output>
  </xyz>
</rpc-reply>
"""
        rpc_reply = NCElement(xmldata, normalize_xslt.encode("UTF-8"))
        self.assertEqual(str(rpc_reply), xmldata_without_ns)

    def _read_file(self, fname):
        fpath = os.path.join(os.path.dirname(__file__), "rpc-reply", fname)
        with open(fpath) as fp:
            foo = fp.read()
        return foo
